# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb.pyfdb_type import UserInputMapper


def test_single_value_internal():
    mars_selection = UserInputMapper.map_selection_to_internal({"key-1": "value-1"})

    assert len(mars_selection) == 1
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-1"]


def test_single_value_external():
    mars_selection = UserInputMapper.map_selection_to_external({"key-1": "value-1"})

    assert len(mars_selection) == 1
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == "value-1"


def test_multi_value_internal():
    mars_selection = UserInputMapper.map_selection_to_internal(
        {
            "key-1": ["value-1", "value-2", "value-3"],
            "key-2": ["value-2"],
            "key-3": ["value-3", 214, 213.54],
            "key-4": [120, 123, 124, 125],
        }
    )

    assert len(mars_selection) == 4
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-1", "value-2", "value-3"]
    assert "key-2" in mars_selection
    assert mars_selection["key-2"] == ["value-2"]
    assert "key-3" in mars_selection
    assert mars_selection["key-3"] == ["value-3", "214", "213.54"]
    assert "key-4" in mars_selection
    assert mars_selection["key-4"] == ["120", "123", "124", "125"]


def test_multi_value_external():
    mars_selection = UserInputMapper.map_selection_to_external(
        {
            "key-1": ["value-1", "value-2", "value-3"],
            "key-2": ["value-2"],
            "key-3": ["value-3", "214", "213.54"],
            "key-4": ["120", "123", "124", "125"],
        }
    )

    assert len(mars_selection) == 4
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-1", "value-2", "value-3"]
    assert "key-2" in mars_selection
    assert mars_selection["key-2"] == ["value-2"]
    assert "key-3" in mars_selection
    assert mars_selection["key-3"] == ["value-3", "214", "213.54"]
    assert "key-4" in mars_selection
    assert mars_selection["key-4"] == ["120", "123", "124", "125"]


def test_to_internal():
    mars_selection = UserInputMapper.map_selection_to_internal(
        {
            "key-1": ["value-2", "value-4"],
            "key-2": ["0.1", 0.2],  # Mixed
            "key-3": [0.1, 0.2],  # Floating
            "key-4": [1, 2],  # Integer
        }
    )

    assert len(mars_selection) == 4
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-2", "value-4"]
    assert "key-2" in mars_selection
    assert mars_selection["key-2"] == ["0.1", "0.2"]
    assert "key-3" in mars_selection
    assert mars_selection["key-3"] == ["0.1", "0.2"]
    assert "key-4" in mars_selection
    assert mars_selection["key-4"] == ["1", "2"]


def test_to_external():
    mars_selection = UserInputMapper.map_selection_to_external(
        {
            "key-1": ["value-2", "value-4"],
            "key-2": ["0.1", "0.2"],  # Mixed
            "key-3": ["0.1", "0.2"],  # Floating
            "key-4": ["1", "2"],  # Integer
        }
    )

    assert len(mars_selection) == 4
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-2", "value-4"]
    assert "key-2" in mars_selection
    assert mars_selection["key-2"] == ["0.1", "0.2"]
    assert "key-3" in mars_selection
    assert mars_selection["key-3"] == ["0.1", "0.2"]
    assert "key-4" in mars_selection
    assert mars_selection["key-4"] == ["1", "2"]


def test_overwrite_key_internal():
    mars_selection = UserInputMapper.map_selection_to_internal(
        {
            "key-1": "value-1",
            "key-1": ["value-3", "214", "213.54"],
        }
    )

    assert len(mars_selection) == 1
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-3", "214", "213.54"]


def test_overwrite_key_external():
    mars_selection = UserInputMapper.map_selection_to_external(
        {
            "key-1": "value-1",
            "key-1": ["value-3", "214", "213.54"],
        }
    )

    assert len(mars_selection) == 1
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-3", "214", "213.54"]


def test_pythonic_interface():
    mars_selection = UserInputMapper.map_selection_to_internal(
        {
            "key-1": ["value-1", "value-2", "value-3"],
            "key-2": 0.1,
            "key-3": list(range(1, 5)),
            "key-4": [0.1, "0.2"],
            "key-5": [1 + 0.5 * x for x in range(2)],
        }
    )

    assert len(mars_selection) == 5
    assert "key-1" in mars_selection
    assert mars_selection["key-1"] == ["value-1", "value-2", "value-3"]
    assert "key-2" in mars_selection
    assert mars_selection["key-2"] == ["0.1"]
    assert "key-3" in mars_selection
    assert mars_selection["key-3"] == ["1", "2", "3", "4"]
    assert "key-4" in mars_selection
    assert mars_selection["key-4"] == ["0.1", "0.2"]
    assert "key-5" in mars_selection
    assert mars_selection["key-5"] == ["1.0", "1.5"]



import logging

import pytest

from pychunked_data_view.exceptions import MarsRequestFormattingError
from z3fdb import (
    AxisDefinition,
    Chunking,
    ExtractorType,
    SimpleStoreBuilder,
)

logging.basicConfig(level=logging.DEBUG)


def test_additional_comma_end_of_request(
    read_only_fdb_pattern_setup,
) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "time=18/0/12/6,"
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166,",  # Ending in an additional comma
        [
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )

    with pytest.raises(MarsRequestFormattingError):
        builder.build()


def test_missing_comma_between_keys(
    read_only_fdb_pattern_setup,
) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "time=18/0/12/6"  # Missing comma here
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166",
        [
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )

    with pytest.raises(MarsRequestFormattingError):
        builder.build()


@pytest.mark.parametrize("invalid_axis", [3, 4])
def test_extend_on_invalid_axis_raises(
    read_only_fdb_pattern_setup,
    invalid_axis,
) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # Three axis definitions ⇒ valid extension axis indices are 0..2.
    # Index 3 is the implicit field dimension (must not be used as extension
    # axis); index 4 is strictly out of bounds.
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "time=0/6/12/18,"
        "date=2020-01-01/to/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167",
        [
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(invalid_axis)

    # pybind11's default translation of eckit::UserError surfaces as
    # RuntimeError, carrying the C++ exception's what() string.
    with pytest.raises(RuntimeError, match="not referring to a valid axis"):
        builder.build()


def test_wrong_key(
    read_only_fdb_pattern_setup,
) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "blubb=0001,"  # There is no blubb key
        "stream=oper,"
        "time=18/0/12/6,"
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166",
        [
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )

    with pytest.raises(MarsRequestFormattingError):
        builder.build()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import sys
from pathlib import Path
from unittest.mock import MagicMock

import findlibs
import pyfdb._internal as _internal
import pytest

from pyfdb.__main__ import DEPENDENCY_ORDER, OPTIONAL_DEPENDENCIES, main


ALL_LIBS = {
    "fdb5": "/fake/fdb/lib/libfdb5.so",
    "eckit": "/fake/eckit/lib/libeckit.so",
    "metkit": "/fake/metkit/lib/libmetkit.so",
    "eccodes": "/fake/eccodes/lib/libeccodes.so",
}

FAKE_VERSION_INFO = [
    ("fdb", "5.13.0", "abc1234", "/fake/fdb/lib/libfdb5.so"),
    ("eckit", "1.32.5", "def5678", "/fake/eckit/lib/libeckit.so"),
    ("metkit", "1.15.0", "ghi9012", "/fake/metkit/lib/libmetkit.so"),
    ("eccodes", "2.10.0", "jkl3456", "/fake/eccodes/lib/libeccodes.so"),
]


def _run_cli(args, monkeypatch):
    monkeypatch.setattr(sys, "argv", ["pyfdb"] + args)
    try:
        main()
        return 0
    except SystemExit as exc:
        return exc.code


def _expected_entry(name, path):
    """Build the expected log fragment for a findlibs lookup line."""
    label = f"{name} [Optional]" if name in OPTIONAL_DEPENDENCIES else name
    home = Path(path).parent.parent
    return f"{label}: {home}"


@pytest.fixture(autouse=True)
def capture_info_logs(caplog):
    caplog.set_level(logging.INFO)


@pytest.fixture(autouse=True)
def version_info_mock(monkeypatch):
    mock = MagicMock(return_value=FAKE_VERSION_INFO)
    monkeypatch.setattr(_internal, "version_info", mock)
    return mock


@pytest.fixture
def find_mock(monkeypatch):
    mock = MagicMock(side_effect=lambda name: ALL_LIBS.get(name))
    monkeypatch.setattr(findlibs, "find", mock)
    return mock


# ---------------------------------------------------------------------------
# --print-home
# ---------------------------------------------------------------------------


def test_print_home_success(find_mock, monkeypatch, caplog):
    exit_code = _run_cli(["--print-home"], monkeypatch)
    assert exit_code == 0


def test_print_home_logs_version_info(find_mock, monkeypatch, caplog):
    _run_cli(["--print-home"], monkeypatch)
    name, version, git_sha, path = FAKE_VERSION_INFO[0]  # fdb entry
    assert name in caplog.text
    assert version in caplog.text
    assert git_sha in caplog.text


def test_print_home_not_found(monkeypatch, caplog):
    monkeypatch.setattr(findlibs, "find", MagicMock(return_value=None))
    exit_code = _run_cli(["--print-home"], monkeypatch)
    assert exit_code == 1
    assert "not found by findlibs" in caplog.text


def test_print_home_calls_find_with_fdb5(find_mock, monkeypatch):
    _run_cli(["--print-home"], monkeypatch)
    find_mock.assert_called_once_with("fdb5")


# ---------------------------------------------------------------------------
# --print-home-deps
# ---------------------------------------------------------------------------


def test_print_home_deps_all_found(find_mock, monkeypatch, caplog):
    exit_code = _run_cli(["--print-home-deps"], monkeypatch)
    assert exit_code == 0
    for name, path in ALL_LIBS.items():
        assert _expected_entry(name, path) in caplog.text


def test_print_home_deps_logs_dependency_versions(find_mock, monkeypatch, caplog):
    _run_cli(["--print-home-deps"], monkeypatch)
    assert "Dependency Versions:" in caplog.text
    for name, version, git_sha, _ in FAKE_VERSION_INFO:
        assert name in caplog.text
        assert version in caplog.text
        assert git_sha in caplog.text


def test_print_home_deps_dependency_versions_logged_before_exit(monkeypatch, caplog):
    libs_without_metkit = {k: v for k, v in ALL_LIBS.items() if k != "metkit"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_metkit.get(n))
    )
    _run_cli(["--print-home-deps"], monkeypatch)
    assert "Dependency Versions:" in caplog.text
    for name, version, git_sha, _ in FAKE_VERSION_INFO:
        assert name in caplog.text
        assert version in caplog.text
        assert git_sha in caplog.text


def test_print_home_deps_missing_required_exits_nonzero(monkeypatch, caplog):
    libs_without_metkit = {k: v for k, v in ALL_LIBS.items() if k != "metkit"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_metkit.get(n))
    )
    exit_code = _run_cli(["--print-home-deps"], monkeypatch)
    assert exit_code == 1
    assert "metkit" in caplog.text


def test_print_home_deps_missing_required_logs_error(monkeypatch, caplog):
    libs_without_metkit = {k: v for k, v in ALL_LIBS.items() if k != "metkit"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_metkit.get(n))
    )
    _run_cli(["--print-home-deps"], monkeypatch)
    errors = [
        r for r in caplog.records
        if r.levelno == logging.ERROR and "metkit" in r.message
    ]
    assert errors


def test_print_home_deps_missing_optional_exits_nonzero(monkeypatch):
    libs_without_eccodes = {k: v for k, v in ALL_LIBS.items() if k != "eccodes"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_eccodes.get(n))
    )
    exit_code = _run_cli(["--print-home-deps"], monkeypatch)
    assert exit_code == 1


def test_print_home_deps_missing_optional_logs_info_not_error(monkeypatch, caplog):
    libs_without_eccodes = {k: v for k, v in ALL_LIBS.items() if k != "eccodes"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_eccodes.get(n))
    )
    _run_cli(["--print-home-deps"], monkeypatch)
    eccodes_records = [r for r in caplog.records if "eccodes" in r.message]
    assert eccodes_records
    assert all(r.levelno == logging.INFO for r in eccodes_records)


def test_print_home_deps_optional_marker_in_message(monkeypatch, caplog):
    libs_without_eccodes = {k: v for k, v in ALL_LIBS.items() if k != "eccodes"}
    monkeypatch.setattr(
        findlibs, "find", MagicMock(side_effect=lambda n: libs_without_eccodes.get(n))
    )
    _run_cli(["--print-home-deps"], monkeypatch)
    assert "[Optional]" in caplog.text


def test_print_home_deps_queries_all_deps(find_mock, monkeypatch):
    _run_cli(["--print-home-deps"], monkeypatch)
    queried = {call.args[0] for call in find_mock.call_args_list}
    assert queried == set(DEPENDENCY_ORDER)


def test_print_home_deps_disable_vars_appear_before_homes(
    find_mock, monkeypatch, caplog
):
    monkeypatch.setenv("FINDLIBS_DISABLE_FDB5", "1")
    _run_cli(["--print-home-deps"], monkeypatch)
    lines = caplog.text.splitlines()
    disable_idx = next(
        i for i, line in enumerate(lines) if "FINDLIBS_DISABLE_FDB5" in line
    )
    first_dep_idx = next(i for i, line in enumerate(lines) if "fdb5:" in line)
    assert disable_idx < first_dep_idx


# ---------------------------------------------------------------------------
# Output format
# ---------------------------------------------------------------------------


def test_logging_format(monkeypatch):
    calls = []
    monkeypatch.setattr(logging, "basicConfig", lambda **kwargs: calls.append(kwargs))
    monkeypatch.setattr(
        findlibs, "find", MagicMock(return_value="/fake/fdb/lib/libfdb5.so")
    )
    _run_cli(["--print-home"], monkeypatch)
    assert calls, "basicConfig must be called"
    fmt = calls[0]["format"]
    assert "%(asctime)s" in fmt
    assert "%(levelname)" in fmt
    assert "%(message)s" in fmt


def test_verbose_sets_debug_level(monkeypatch):
    calls = []
    monkeypatch.setattr(logging, "basicConfig", lambda **kwargs: calls.append(kwargs))
    monkeypatch.setattr(
        findlibs, "find", MagicMock(return_value="/fake/fdb/lib/libfdb5.so")
    )
    _run_cli(["--print-home", "--verbose"], monkeypatch)
    assert calls[0]["level"] == logging.DEBUG


def test_default_level_is_info(monkeypatch):
    calls = []
    monkeypatch.setattr(logging, "basicConfig", lambda **kwargs: calls.append(kwargs))
    monkeypatch.setattr(
        findlibs, "find", MagicMock(return_value="/fake/fdb/lib/libfdb5.so")
    )
    _run_cli(["--print-home"], monkeypatch)
    assert calls[0]["level"] == logging.INFO


# ---------------------------------------------------------------------------
# No arguments
# ---------------------------------------------------------------------------


def test_no_args_exits_with_code_2(monkeypatch):
    exit_code = _run_cli([], monkeypatch)
    assert exit_code == 2



# Configuration file for the Sphinx documentation builder.
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# NOTE: This configuration uses autoapi without automatic generation and
# instead uses directives (see api.rst). This is done because with just
# autoapi auto generation all modules must be imported. In ource case this
# would mean we need to complie our python bindings. Using autoapi with
# directives allows to use 'static analysis' mode to extract docstrings.default_mod
# See: https://sphinx-autoapi.readthedocs.io/en/stable/how_to.html#how-to-transition-to-autodoc-style-documentation
# See: https://sphinx-autoapi.readthedocs.io/en/stable/reference/directives.html

import datetime

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


project = "z3FDB"
copyright = f"{datetime.datetime.today().year}, ECMWF"
author = "ECMWF"
version = "local-dev"
release = "local-dev"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.mermaid",
    "autoapi.extension",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

# -- Autoapi settings
autoapi_dirs = ["../../src/z3fdb", "../../src/pychunked_data_view"]
autoapi_type = "python"
autoapi_generate_api_docs = False
autoapi_add_toctree_entry = False
autoapi_python_class_content = "class"
autoapi_ignore = [
    "*/_internal/*",
]
add_module_names = False

# -- Napoleon settings
napoleon_google = True

templates_path = ["_templates"]
exclude_patterns = ["Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
# Do not show source link on each page
html_show_sourcelink = False
html_sidebars = {
    "**": [],
}
html_theme_options = {
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["navbar-icon-links", "theme-switcher", "version-switcher"],
    "navbar_persistent": ["search-button"],
    "primary_sidebar_end": [],
    # On local builds no version.json is present
    "check_switcher": False,
}

html_context = {
    # Enable auto detection of light/dark mode
    "default_mode": "auto"
}

# -- autosectionlabel configuration ------------------------------------------
autosectionlabel_prefix_document = True



import pytest

from pyfdb.pyfdb import FDB
from pyfdb.pyfdb_iterator import IndexAxis


def test_index_axis_not_implemented():
    selection = {
        "type": ["an"],
        "class": ["ea"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
        "date": ["20200101"],
        "levtype": ["sfc"],
        "step": ["0"],
        "param": ["167", "131", "132"],
        "time": ["1800"],
    }

    index_axis = IndexAxis(selection)

    with pytest.raises(TypeError):
        index_axis["date"] = "1999"

    with pytest.raises(TypeError):
        del index_axis["date"]

    with pytest.raises(TypeError):
        index_axis.clear()


def test_index_axis_creation_dict():
    selection = {
        "type": ["an"],
        "class": ["ea"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
        "date": ["20200101"],
        "levtype": ["sfc"],
        "step": ["0"],
        "param": ["167", "131", "132"],
        "time": ["1800"],
    }

    index_axis = IndexAxis(selection)

    assert index_axis.items() == selection.items()
    for value in selection.values():
        assert value in index_axis.values()
    assert index_axis.keys() == selection.keys()
    assert index_axis == selection


def test_index_axis_creation_dict_no_collections():
    selection = {
        "type": ["an"],
        "class": ["ea"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
        "date": ["20200101"],
        "levtype": ["sfc"],
        "step": ["0"],
        "param": ["167", "131", "132"],
        "time": ["1800"],
    }

    index_axis = IndexAxis(selection)

    non_collection_selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "131", "132"],
        "time": "1800",
    }

    expected = IndexAxis(non_collection_selection)

    assert expected.items() == expected.items()
    for value in expected.values():
        assert value in index_axis.values()
    assert expected.keys() == expected.keys()
    assert index_axis == expected


def test_index_axis_with_dict(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": ["an"],
        "class": ["ea"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
        "date": ["20200101"],
        "levtype": ["sfc"],
        "step": ["0"],
        "param": ["167", "131", "132"],
        "time": ["1800"],
    }

    index_axis = fdb.axes(selection)

    expected_index_axis = {
        "type": ["an"],
        "class": ["ea"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
        "date": ["20200101"],
        "levelist": [""],
        "levtype": ["sfc"],
        "step": ["0"],
        "param": ["131", "132", "167"],
        "time": ["1800"],
    }

    assert index_axis == expected_index_axis
    assert expected_index_axis == index_axis


def test_index_axis_string(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis = fdb.axes(selection)

    assert index_axis
    assert str(index_axis)


def test_index_axis_get(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert index_axis
    assert index_axis["class"]
    assert index_axis["domain"]
    assert index_axis["date"]
    assert index_axis["expver"]
    assert index_axis["stream"]
    assert index_axis["time"]

    with pytest.raises(KeyError, match="Couldn't find key:"):
        assert index_axis["non-existing-key"]


def test_fdb_index_axis_in(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert index_axis
    assert len(index_axis.keys()) == 11

    for key in [
        "class",
        "date",
        "domain",
        "expver",
        "levelist",
        "levtype",
        "param",
        "step",
        "stream",
        "time",
        "type",
    ]:
        assert key in index_axis

    assert index_axis["class"] == ["ea"]
    assert index_axis["date"] == ["20200101"]
    assert index_axis["domain"] == ["g"]
    assert index_axis["expver"] == ["0001"]
    assert index_axis["levelist"] == [""]
    assert index_axis["levtype"] == ["sfc"]
    assert index_axis["param"] == ["131", "132", "167"]
    assert index_axis["step"] == ["0"]
    assert index_axis["stream"] == ["oper"]
    assert index_axis["time"] == ["1800"]
    assert index_axis["type"] == ["an"]

    assert "non-existing-key" not in index_axis


def test_index_axis_repr(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert repr(index_axis)


def test_index_axis_keys(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert len(index_axis.keys()) == 11


def test_index_axis_values(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert len(index_axis) == 11


def test_index_axis_items(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert len(index_axis) == 11
    assert len(index_axis.items()) == 11
    assert len(index_axis.keys()) == 11
    assert len(index_axis.values()) == 11

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")


def test_index_axis_items_levels(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        # "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    index_axis: IndexAxis = fdb.axes(selection)

    assert len(index_axis) == 11

    print("---------- Level 3: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")

    for k, v in index_axis.items():
        assert index_axis.has_key(k)

    index_axis: IndexAxis = fdb.axes(selection, level=2)

    assert len(index_axis) == 8

    print("---------- Level 2: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")

    index_axis: IndexAxis = fdb.axes(selection, level=1)

    assert len(index_axis) == 6

    print("---------- Level 1: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")


def test_index_axis_items_empty_request(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    index_axis: IndexAxis = fdb.axes({})

    assert len(index_axis) == 11

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")

    for k, v in index_axis.items():
        assert index_axis.has_key(k)


def test_index_axis_equality(read_only_fdb_setup):
    with FDB(read_only_fdb_setup) as fdb:
        selection = {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "131/132/167",
            "time": "1800",
        }

        index_axis_1: IndexAxis = fdb.axes(selection)
        index_axis_2: IndexAxis = fdb.axes(selection)

        assert index_axis_1 is not index_axis_2
        assert index_axis_1 == index_axis_2

        # Check for inequality if different axis
        index_axis_non_existent: IndexAxis = fdb.axes(
            {
                "type": "an",
                "class": "ea",
                "domain": "g",
                "expver": "0001",
                "stream": "oper",
                "date": "20200101",
                "levtype": "sfc",
                "step": "0",
                "param": "131/132/167",
                "time": "1900",  # This differs
            }
        )

        assert index_axis_1 is not index_axis_non_existent
        assert index_axis_1 != index_axis_non_existent


def test_index_axis_equality_dict(read_only_fdb_setup):
    with FDB(read_only_fdb_setup) as fdb:
        selection = {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "131/132/167",
            "time": "1800",
        }

        index_axis_1: IndexAxis = fdb.axes(selection)

        index_axis_manual = {
            "class": ["ea"],
            "domain": ["g"],
            "expver": "0001",
            "stream": ["oper"],
            "date": ["20200101"],
            "time": ["1800"],
            "levelist": "",
            "levtype": ["sfc"],
            "param": ["131", "132", "167"],
            "step": "0",
            "type": "an",
        }

        assert index_axis_1 is not index_axis_manual
        assert index_axis_1 == index_axis_manual

        index_axis_manual_non_existing = {
            "class": ["ea"],
            "domain": ["g"],
            "expver": "0001",
            "stream": ["oper"],
            "date": ["20200101"],
            "time": ["1900"],  # This differs
            "levelist": "",
            "levtype": ["sfc"],
            "param": ["131", "132", "167"],
            "step": "0",
            "type": "an",
        }

        assert index_axis_1 is not index_axis_manual_non_existing
        assert index_axis_1 != index_axis_manual_non_existing



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
import numpy as np
import itertools


from pychunked_data_view import (
    AxisDefinition,
    ChunkedDataViewBuilder,
    Chunking,
    ExtractorType,
)


def test_axis_definition_can_construct():
    obj = AxisDefinition(["key1", "key0"], Chunking.SINGLE_VALUE)
    assert obj.keys == ["key1", "key0"]
    assert obj.chunking == Chunking.SINGLE_VALUE


def test_axis_definition_can_assign():
    obj = AxisDefinition(["key1", "key0"], Chunking.SINGLE_VALUE)
    assert obj.keys == ["key1", "key0"]
    assert obj.chunking == Chunking.SINGLE_VALUE
    obj.keys = []
    assert obj.keys == []
    obj.chunking = Chunking.SINGLE_VALUE
    assert obj.chunking == Chunking.SINGLE_VALUE


def test_builder(read_only_fdb_setup):
    builder = ChunkedDataViewBuilder(read_only_fdb_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-04,"
        "levtype=sfc,"
        "step=0,"
        "param=167/131/132,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    view = builder.build()

    expected = list(range(0, 5248))
    for a, b in itertools.product(range(0, 32), range(0, 3)):
        np.testing.assert_array_almost_equal_nulp(view.at((a, b, 0)), expected)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import datetime
import os
# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


project = "Fields DataBase - FDB"
copyright = f"{datetime.datetime.today().year}, ECMWF"
author = "ECMWF"
version = "local-dev"
release = "local-dev"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.mermaid",
    "autoapi.extension",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
    "breathe",
    "sphinx.ext.autosectionlabel",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "_internal"]

autoapi_dirs = ["../src/pyfdb", "../src/z3fdb", "../src/pychunked_data_view"]
autoapi_type = "python"
autoapi_generate_api_docs = True
autoapi_add_toctree_entry = False
autoapi_python_class_content = "both"
autoapi_ignore = [
    "*/_internal/*",
]
add_module_names = False
autoapi_keep_files = False

# -- Napoleon settings
napoleon_google = True

templates_path = ["_templates"]
exclude_patterns = ["Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_context = {"default_mode": "auto"}
html_theme_options = {
    "switcher": {
        "json_url": "https://sites.ecmwf.int/docs/fdb/versions.json",
        "version_match": version,
    },
    "show_toc_level": 2,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/ecmwf/fdb",
            "icon": "fa-brands fa-github",
        },
    ],
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["navbar-icon-links", "theme-switcher", "version-switcher"],
    "navbar_persistent": ["search-button"],
    "secondary_sidebar_items": ["page-toc", "edit-this-page", "sourcelink"],
    # On local builds no version.json is present
    "check_switcher": False,
    "content_footer_items": [],
    "footer_start": ["copyright"],
    "footer_center": [],
    "footer_end": [],
}
html_sidebars = {"**": ["sidebar-nav-bs"]}
html_static_path = ["_static"]

# -- Breathe configuration ---------------------------------------------------
breathe_projects = {"FDB": os.environ.get("DOXYGEN_XML_DIR", "doxygen/xml")}
breathe_default_project = "FDB"

# -- autosectionlabel configuration ------------------------------------------
autosectionlabel_prefix_document = True



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb import FDB
from pyfdb.pyfdb_iterator import ListElement


def test_inspect(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131",
        "time": "1800",
    }

    list_iterator = fdb.inspect(selection)

    assert list_iterator

    elements: list[ListElement] = []

    for el in list_iterator:
        elements.append(el)
        print(el)

    assert len(elements) == 1

    for el in elements:
        data_handle = el.data_handle
        assert data_handle
        data_handle.open()
        assert data_handle.read(4) == b"GRIB"
        data_handle.close()


def test_inspect_multiple_values(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/167",
        "time": "1800",
    }

    list_iterator = fdb.inspect(selection)

    assert list_iterator

    elements = []

    for el in list_iterator:
        elements.append(el)
        print(el)

    assert len(elements) == 2

    for el in elements:
        data_handle = el.data_handle
        assert data_handle
        data_handle.open()
        assert data_handle.read(4) == b"GRIB"
        data_handle.close()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb import FDB, URI


def test_list_element_uri(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=1)
    assert list_iterator

    elements = list(list_iterator)
    assert len(elements) == 1
    print(elements[0])
    print(elements[0].uri)

    print("----------------------------------")

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=2)
    assert list_iterator

    elements = list(list_iterator)
    assert len(elements) == 1
    print(elements[0])
    print(elements[0].uri)

    print("----------------------------------")

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=3)
    assert list_iterator

    elements = [el.uri for el in list_iterator if el.uri]

    assert len(elements) == 3
    for el in elements:
        assert el is not None
        assert el.path() is not None


def test_uri_roundtrip():
    uri_str = "scheme://user:secretpass@example.com:8443/path/to/resource?query=search&sort=asc#section-2"
    test_uri = URI(uri_str)

    assert test_uri.scheme() == "scheme"
    assert test_uri.username() == "user"
    assert test_uri.password() == "secretpass"
    assert test_uri.hostname() == "example.com"
    assert test_uri.hostname() == "example.com"
    assert test_uri.netloc() == "user:secretpass@example.com:8443"
    assert test_uri.port() == 8443
    assert test_uri.path() == "/path/to/resource"
    assert test_uri.query() == "query=search&sort=asc"
    assert test_uri.fragment() == "section-2"

    assert str(test_uri) == uri_str

    print(test_uri)


def test_uri_cmp():
    uri_str = "scheme://user:secretpass@example.com:8443/path/to/resource?query=search&sort=asc#section-2"
    test_uri = URI(uri_str)

    uri_str_section3 = "scheme://user:secretpass@example.com:8443/path/to/resource?query=search&sort=asc#section-3"

    test_uri_section3 = URI(uri_str_section3)

    assert test_uri != test_uri_section3
    assert uri_str_section3 != test_uri_section3



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import pytest
import zarr
import numpy as np

from z3fdb import (
    AxisDefinition,
    Chunking,
    SimpleStoreBuilder,
    ExtractorType,
)


def test_zarr_use_spec_v2(read_only_fdb_setup) -> None:
    assert zarr.config.get("default_zarr_format") == 3


def test_access(read_only_fdb_setup) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-04,"
        "levtype=sfc,"
        "step=0,"
        "param=167/131/132,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data
    print(data[:, :])
    print(data.shape)


def test_axis_check_out_of_bounds(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """This test checks whether an access to an axis which has no pendant in the data is failing.
    The request below has param 167 which is not given in the data of the setup fdb. Therefore this
    needs to fail. Accessing the first two params is fine.
    """
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    for i in range(16):
        assert np.all(data[i, 0] == 2 * i)
        assert np.all(data[i, 1] == (2 * i) + 1)

    with pytest.raises(Exception):
        data[0, 2]



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb import FDB


def test_stats(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "167/165/166",
        "time": "1800",
    }

    stats_iterator = fdb.stats(selection)
    assert stats_iterator

    elements = []

    for el in stats_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1
    stats = str(elements[0])

    # Check for database and index statistics
    assert "Index Statistics:" in stats
    assert "Fields                          : 3" in stats
    assert "Reacheable fields               : 3" in stats
    assert "DB Statistics:" in stats
    assert "Databases                       : 1" in stats
    assert "TOC records                     : 2" in stats
    assert "TOC records                     : 2" in stats
    assert "Owned data files                : 1" in stats
    assert "Index files                     : 1" in stats

    # Check fields
    assert "Fields" in stats
    assert "Size of fields" in stats
    assert "Reacheable fields" in stats
    assert "Reachable size" in stats
    assert "Databases" in stats
    assert "TOC records" in stats
    assert "Size of TOC files" in stats
    assert "Size of schemas files" in stats
    assert "TOC records" in stats
    assert "Owned data files" in stats
    assert "Size of owned data files" in stats
    assert "Index files" in stats
    assert "Size of index files" in stats
    assert "Size of TOC files" in stats
    assert "Total owned size" in stats
    assert "Total size" in stats


def test_stats_db_stats(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "167/165/166",
        "time": "1800",
    }

    stats_iterator = fdb.stats(selection)
    assert stats_iterator

    elements = list(stats_iterator)

    assert len(elements) == 1
    stats = elements[0].db_statistics()

    print(stats)

    # Check for database and index statistics
    assert "Index Statistics:" not in stats
    assert "Fields                          : 3" not in stats
    assert "Reacheable fields               : 3" not in stats
    assert "DB Statistics:" not in stats
    assert "Databases                       : 1" in stats
    assert "TOC records                     : 2" in stats
    assert "TOC records                     : 2" in stats
    assert "Owned data files                : 1" in stats
    assert "Index files                     : 1" in stats

    # Check fields
    assert "TOC records" in stats
    assert "Size of TOC files" in stats
    assert "Size of schemas files" in stats
    assert "TOC records" in stats
    assert "Owned data files" in stats
    assert "Size of owned data files" in stats
    assert "Index files" in stats
    assert "Size of index files" in stats
    assert "Size of TOC files" in stats
    assert "Total owned size" in stats
    assert "Total size" in stats


def test_stats_index_stats(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "167/165/166",
        "time": "1800",
    }

    stats_iterator = fdb.stats(selection)
    assert stats_iterator

    elements = list(stats_iterator)

    assert len(elements) == 1
    stats = elements[0].index_statistics()

    print(stats)

    # Check for database and index statistics
    assert "Fields                          : 3" in stats
    assert "Reacheable fields               : 3" in stats
    assert "DB Statistics:" not in stats
    assert "Databases                       : 1" not in stats
    assert "TOC records                     : 2" not in stats
    assert "TOC records                     : 2" not in stats
    assert "Owned data files                : 1" not in stats
    assert "Index files                     : 1" not in stats

    # Check fields
    assert "Fields" in stats
    assert "Size of fields" in stats
    assert "Reacheable fields" in stats
    assert "Reachable size" in stats



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""Top-level pytest configuration."""

import itertools
import pathlib
import shutil

import eccodes as ec
import pytest
import yaml
from numpy import repeat

from pyfdb import FDB


def create_fdb(root: pathlib.Path, schema_src: pathlib.Path) -> pathlib.Path:
    """Create a new FDB under 'root' using the provided schema.

    This helper function is supposed to be used to build specific test FDB setups.

    Args:
        root (pathlib.Path): Directory that will hold the resulting ``.fdb``.
        schema_src (pathlib.Path): Path to a text file containing the schema definition.

    Returns:
        pathlib.Path: Absolute path to the created ``fdb_config.yaml`` file.

    Asserts:
        - ``root`` needs to already exist.
        - ``schema`` file exists
    """
    assert root.is_dir()
    assert schema_src.is_file()
    schema_path = root / "schema"
    shutil.copy(schema_src, schema_path)

    db_store_path = root / "db_store"
    db_store_path.mkdir()
    fdb_config = dict(
        type="local",
        engine="toc",
        schema=str(schema_path),
        spaces=[
            dict(
                handler="Default",
                roots=[
                    {"path": str(db_store_path)},
                ],
            )
        ],
    )
    fdb_config_str = yaml.dump(fdb_config)
    fdb_config_path = root / "fdb_config.yaml"
    fdb_config_path.write_text(fdb_config_str)
    return fdb_config_path


def populate_fdb(config: pathlib.Path, messages: list[pathlib.Path]):
    """Import 'messages' into fdb pointed to by 'config'

    This helper function is supposed to be used to build specific test FDB setups.

    Args:
        config (pathlib.Path): FDB database to use.
        messages (list[pathlib.Path]): List of files with messages to ingest.

    Asserts:
        - FDB config file needs to exist
        - All message files need to exist

    """
    assert config.is_file()
    fdb = FDB(config.read_text())
    for message in messages:
        assert message.is_file()
        fdb.archive(message.read_bytes())
    fdb.flush()


@pytest.fixture(scope="session")
def data_path() -> pathlib.Path:
    """
    Provides path to test data
    """
    path = pathlib.Path(__file__).resolve().parent / "data"
    assert path.exists()
    return path


@pytest.fixture(scope="session")
def session_tmp(tmp_path_factory) -> pathlib.Path:
    return tmp_path_factory.mktemp("session_data")


@pytest.fixture(scope="session")
def build_example_sfc_pl_grib_messages(data_path, session_tmp) -> pathlib.Path:
    """
    Build messages that span the data required to match mars requests in
    'data/anemoi-recipes/example.yaml'
    """
    tmp = session_tmp / "build_example_sfc_pl_messages"
    tmp.mkdir()
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")
    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()
    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing datapoints
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")

    dates = [20200101, 20200102]
    times = [0, 300, 600, 900, 1200, 1500, 1800, 2100]
    # 10u/10v
    parameters_sfc = [165, 166]

    # u/v
    parameters_pl = [131, 132]
    levels = [50, 100]

    messages = tmp / "test_data.grib"
    with open(messages, "wb") as out:
        ec.codes_set_string(gid, "levtype", "sfc")
        for value, (date, time, parameter) in enumerate(
            itertools.product(dates, times, parameters_sfc)
        ):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)

            ec.codes_set_values(gid, repeat(value, count_values))
            ec.codes_write(gid, out)

        offset = value

        ec.codes_set_string(gid, "levtype", "pl")
        for value, (date, time, level, parameter) in enumerate(
            itertools.product(dates, times, levels, parameters_pl)
        ):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_set(gid, "level", level)
            ec.codes_set_values(gid, repeat(offset + value, count_values))
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages


@pytest.fixture(scope="session")
def build_grib_messages(data_path, session_tmp) -> pathlib.Path:
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")
    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()
    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing datapoints
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")
    ec.codes_set_string(gid, "levtype", "sfc")
    ec.codes_set_values(gid, list(range(0, count_values)))

    dates = [20200101, 20200102, 20200103, 20200104]
    times = [0, 300, 600, 900, 1200, 1500, 1800, 2100]
    parameters = [167, 131, 132]

    messages = session_tmp / "test_data.grib"
    with open(messages, "wb") as out:
        for date, time, parameter in itertools.product(dates, times, parameters):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages


@pytest.fixture(scope="session", autouse=False)
def read_only_fdb_setup_for_sfc_pl_example(
    data_path, session_tmp, build_example_sfc_pl_grib_messages
) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    Test FDB currently reads all grib files in `tests/data`
    This setup can be shared between tests as we will only read
    data from this FDB
    """
    fdb_root = session_tmp / "sfc-pl-example-fdb"
    fdb_root.mkdir()
    schema_path_src = data_path / "schema"
    cfg = create_fdb(fdb_root, schema_path_src)
    populate_fdb(cfg, [build_example_sfc_pl_grib_messages])
    return cfg


@pytest.fixture(scope="session", autouse=False)
def read_only_fdb_setup(data_path, session_tmp, build_grib_messages) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    Test FDB currently reads all grib files in `tests/data`
    This setup can be shared between tests as we will only read
    data from this FDB
    """
    schema_path_src = data_path / "schema"
    assert schema_path_src.is_file()
    schema_path = session_tmp / "schema"
    shutil.copy(schema_path_src, schema_path)

    db_store_path = session_tmp / "db_store"
    db_store_path.mkdir()
    fdb_config = dict(
        type="local",
        engine="toc",
        schema=str(schema_path),
        spaces=[
            dict(
                handler="Default",
                roots=[
                    {"path": str(db_store_path)},
                ],
            )
        ],
    )
    fdb_config_str = yaml.dump(fdb_config)
    fdb_config_path = session_tmp / "fdb_config.yaml"
    fdb_config_path.write_text(fdb_config_str)
    fdb = FDB(fdb_config_str)
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()
    return fdb_config_path


@pytest.fixture(scope="session", autouse=False)
def read_only_fdb_pattern_setup(
    data_path, session_tmp, build_pattern_grib_messages
) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    Test FDB currently reads all grib files in `tests/data`
    This setup can be shared between tests as we will only read
    data from this FDB
    """
    schema_path_src = data_path / "schema"
    assert schema_path_src.is_file()
    schema_path = session_tmp / "schema"
    shutil.copy(schema_path_src, schema_path)

    db_store_path = session_tmp / "db_store_pattern"
    db_store_path.mkdir()
    fdb_config = dict(
        type="local",
        engine="toc",
        schema=str(schema_path),
        spaces=[
            dict(
                handler="Default",
                roots=[
                    {"path": str(db_store_path)},
                ],
            )
        ],
    )
    fdb_config_str = yaml.dump(fdb_config)
    fdb_config_path = session_tmp / "fdb_config.yaml"
    fdb_config_path.write_text(fdb_config_str)
    fdb = FDB(fdb_config_str)
    fdb.archive(build_pattern_grib_messages.read_bytes())
    fdb.flush()
    return fdb_config_path


@pytest.fixture(scope="session")
def build_pattern_grib_messages(data_path, session_tmp) -> pathlib.Path:
    """
    Build messages which have a certain pattern to control the correct assembly of
    zarr files
    """
    tmp = session_tmp / "build_pattern_grib_messages"
    tmp.mkdir()
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")

    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()

    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing data points
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")

    dates = [20200101, 20200102, 20200103]
    times = [0000, 600, 1200, 1800]

    # 10u/10v
    parameters_sfc = [165, 166, 167]

    # u/v
    parameters_pl = [131, 132, 133]
    levels = [50, 100, 150]

    messages = tmp / "test_data_pattern.grib"
    with open(messages, "wb") as out:
        ec.codes_set_string(gid, "levtype", "sfc")

        total_values = 0

        for value, (date, time, parameter) in enumerate(
            itertools.product(dates, times, parameters_sfc)
        ):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_set_values(gid, list(itertools.repeat(value, count_values)))
            ec.codes_write(gid, out)

            total_values += 1

        ec.codes_set_string(gid, "levtype", "pl")
        for value, (date, time, parameter, level) in enumerate(
            itertools.product(dates, times, parameters_pl, levels)
        ):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_set(gid, "level", level)
            ec.codes_set_values(gid, list(itertools.repeat(total_values + value, count_values)))
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

from collections.abc import Collection, Mapping
from enum import IntEnum, IntFlag, auto
from pathlib import Path
from urllib import parse

from pyfdb._internal import _URI, _ControlAction, _ControlIdentifier, _DataHandle
from pyfdb._internal.pyfdb_internal import InternalMarsIdentifier, InternalMarsSelection

MarsSelection = Mapping[str, str | int | float | Collection[str | int | float]]
"""
Selection part of a MARS request.

This is a key-value map, with the data types allowed below
"""

MarsIdentifier = list[tuple[str, str]] | dict[str, str]
"""
This is the representation of a MARS identifier

This is a key-value List, mapping MARS keys to a string resembling a singluar value, see https://github.com/ecmwf/datacube-spec.
"""


class UserInputMapper:
    """
    Selection mapper for creating MARS selections.

    This class helps to create syntactically correctly structured MARS selections. If `strict_mode`
    is activated there will be checks whether keys have been set already.

    """

    @classmethod
    def map_selection_to_internal(cls, selection: MarsSelection) -> InternalMarsSelection:
        result: Mapping[str, Collection[str]] = {}

        for key, values in selection.items():
            if not isinstance(values, (int, float, str, Collection)) or isinstance(values, Mapping):
                raise ValueError(
                    f"MarsSelectionMapper: The given value for key '{key}' is not valid. A MarsSelection has to have the following type: {MarsSelection}"
                )

            # Values is a list of values but not a single string
            if isinstance(values, Collection) and not isinstance(values, str):
                converted_values = [str(v) if isinstance(v, (float, int)) else v for v in values]
                result[key] = converted_values
            # Values is a string or a float or an int
            elif isinstance(values, (int, float)):
                result[key] = [str(values)]
            elif isinstance(values, str):
                # Note: Even `to`/`to-by` range expressions are split here and treated as plain values.
                #       UserInputMapper.map_selection_to_internal() is used for all APIs (retrieve,
                #       inspect, axes, etc.). The semantic expansion of these ranges now happens in
                #       the bindings layer (see mars_requestfrom_map()), not directly in the FDB
                #       tool request. We hand a list[str] to the pybind11 layer, matching the C++
                #       splitting behaviour.
                if "/" in values:
                    result[key] = values.split("/")
                else:
                    result[key] = [values]
            else:
                raise ValueError(
                    f"Unknown type for key: {key}. Type must be int, float, str or a collection of those."
                )

        return result

    @classmethod
    def map_selection_to_external(cls, selection: InternalMarsSelection) -> MarsSelection:
        result: MarsSelection = {}

        for key, values in selection.items():
            # Values is a list of values but not a single string
            if isinstance(values, Collection) and not isinstance(values, str):
                converted_values = [str(v) if isinstance(v, (float, int)) else v for v in values]
                result[key] = converted_values
            # Values is a string or a float or an int
            elif isinstance(values, (str, int, float)):
                result[key] = str(values)
            else:
                raise ValueError(
                    f"Unknown type for key: {key}. Type must be int, float, str or a collection of those."
                )

        return result

    @classmethod
    def map_identifier_to_internal(cls, identifier: MarsIdentifier) -> InternalMarsIdentifier:
        key_values: dict[str, str] = {}

        iterator = None

        if isinstance(identifier, list):
            iterator = identifier
        elif isinstance(identifier, dict):
            iterator = identifier.items()
        else:
            raise ValueError(
                "Identifier: Unknown type for key_value_pairs. List[Tuple[str, str]] or Dict[str, str] needed"
            )

        for k, v in iterator:
            if k in key_values:
                raise KeyError(
                    f"Identifier: Key {k} already exists in Identifier: {str(key_values)}"
                )

            if isinstance(v, list) or "/" in v:
                raise ValueError(
                    "No list of values allowed. An Identifier has to be a mapping from a single key to a single value."
                )

            key_values[k] = v

        return list(key_values.items())


class DataHandle:
    """
    DataHandle class for lazy reading from a data source.

    Parameters
    ----------
    None

    Note
    ----
    *This class can't be instantiated / is only returned from the underlying FDB calls*

    Returns
    -------
    `DataHandle` class

    Examples
    --------
    >>> request = {
    >>>     "type": "an",
    >>>     "class": "ea",
    >>>     "domain": "g",
    >>>     "expver": "0001",
    >>>     "stream": "oper",
    >>>     "date": "20200101",
    >>>     "levtype": "sfc",
    >>>     "step": "0",
    >>>     "param": ["167", "165", "166"],
    >>>     "time": "1800",
    >>> }
    >>> data_handle = fdb.retrieve(request)
    >>> data_handle.open()
    >>> data_handle.read(4) == b"GRIB"
    >>> data_handle.close()
    >>> # OR
    >>> with fdb.retrieve(request) as data_handle:
    >>>     data_handle.read(4) == b"GRIB"
    """

    def __init__(self, dataHandle: _DataHandle, *, _internal=False):
        if not _internal:
            raise TypeError("Creating a ListElement from user code is not supported.")
        self.dataHandle: _DataHandle = dataHandle
        self.opened = False

    def __enter__(self) -> "DataHandle":
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()

    def open(self) -> None:
        """
        Open the DataHandle object for reading.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> data_handle = fdb.retrieve(request)
        >>> data_handle.open()
        >>> data_handle.read(4) == b"GRIB"
        >>> data_handle.close()
        """
        if self.opened is True:
            return
        self.opened = True
        self.dataHandle.open()

    def close(self) -> None:
        """
        Close the DataHandle object after reading.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> data_handle = fdb.retrieve(request)
        >>> data_handle.open()
        >>> data_handle.read(4) == b"GRIB"
        >>> data_handle.close()
        """
        if self.opened is False:
            return
        self.opened = False
        self.dataHandle.close()

    def size(self) -> int:
        """
        Return the size of a data handle in bytes.

        Parameters
        ----------
        None

        Returns
        -------
        `int` describing the size of the datahandle in bytes.

        Examples
        --------
        >>> with fdb.retrieve(selection) as data_handle:
        >>>     data_handle.size() # Returns the size of the datahandle in bytes
        """
        return self.dataHandle.size()

    def read(self, len: int = -1) -> bytes:
        """
        Read a given amount of bytes from the `DataHandle`.
        This method copies the data from the underlying memory.

        Parameters
        ----------
        `len`: int
            Number of bytes to read. Defaults to -1, which reads the entire buffer.
            If the requested size exceeds the available data, the result is zero-padded to match the requested size.

        Returns
        -------
        `bytes` object resembling the read data.

        Raises
        ------
        `RuntimeError` if the DataHandle wasn't opened before the read.

        Examples
        --------
        >>> with fdb.retrieve(request) as data_handle
        >>>     data_handle.read(4) == b"GRIB"
        >>>     #or
        >>>     data_handle.read(-1) # Read the entire file
        """
        if self.opened is False:
            raise RuntimeError(
                "DataHandle: Read occured before the handle was opened. Must be opened first."
            )

        # NB: Some data handles do not implement the size() method. In this case, we just read the requested amount of bytes.
        try:
            dh_size = self.dataHandle.size()
            array_size = dh_size if len == -1 else min(len, dh_size)
        except RuntimeError:
            if len == -1:
                raise
            array_size = len

        buffer = bytearray(array_size)
        read_bytes = self.dataHandle.read(buffer)

        if read_bytes == 0:
            return b""

        return bytes(buffer)

    def readinto(self, buffer: memoryview) -> int:
        """
        Read a given amount of bytes from the `DataHandle` into a `memoryview`.
        This is a zero-copy method.

        Parameters
        ----------
        `buffer`: `memoryview`
            Memory view for the buffer in which the bytes should be read

        Returns
        -------
        `int` size of bytes which have been read

        Raises
        ------
        `RuntimeError` if the DataHandle wasn't opened before the read.

        Examples
        --------
        >>> dst_read_into = io.BytesIO(b"")
        >>> with fdb.retrieve(selection) as data_handle:
        >>>     assert data_handle
        >>>     shutil.copyfileobj(data_handle, dst_read_into)
        >>>     # Reset position in file
        >>>     dst_read_into.seek(0)
        """
        if self.opened is False:
            raise RuntimeError(
                "DataHandle: Read occured before the handle was opened. Must be opened first."
            )

        # Buffer is a writable buffer (memoryview)
        return self.dataHandle.read(buffer)

    def readall(self, buffer_size: int = 1024) -> bytes:
        """
        Read all bytes from the DataHandle into memory.
        This method copies the data from the underlying memory.

        Parameters
        ----------
        `buffer_size`: `int`
            The size of the buffer which is used for reading

        Note
        ----
        *There is no need to open the DataHandle before*. This is handled by the function. The
        default chunk size is 1024 bytes.

        Returns
        -------
        `bytes` object resembling the read data

        Examples
        --------
        >>> data_handle = fdb.retrieve(request)
        >>> data_handle.readall() # Returns all content of a datahandle b"GRIB..."
        """
        buffer = bytearray()
        total_bytes_read = 0

        chunk_buf = bytearray(buffer_size)

        self.open()
        bytes_read = self.dataHandle.read(chunk_buf)

        while bytes_read > 0:
            total_bytes_read += bytes_read
            buffer.extend(chunk_buf[:bytes_read])
            bytes_read = self.dataHandle.read(chunk_buf)

        self.close()

        return bytes(buffer)

    def __repr__(self) -> str:
        return f"[{'Opened' if self.opened else 'Closed'}] Datahandle: {self.dataHandle}"


class ControlIdentifier(IntFlag):
    """
    Specify which functionality of the FDB should be addressed, e.g. RETRIEVE or LIST.

    Values
    ------
    - NONE
    - LIST
    - RETRIEVE
    - ARCHIVE
    - WIPE
    - UNIQUEROOT
    """

    NONE = 0
    LIST = auto()
    RETRIEVE = auto()
    ARCHIVE = auto()
    WIPE = auto()
    UNIQUEROOT = auto()

    @classmethod
    def _from_raw(cls, en: _ControlIdentifier):
        return ControlIdentifier[en.name]

    def _to_raw(self):
        return _ControlIdentifier[self.name]

    def __repr__(self) -> str:
        active_identifier = "|".join([val.name for val in ControlIdentifier if self.value & val])
        if len(active_identifier) == 0:
            return "NONE"

        return active_identifier

    def __str__(self) -> str:
        return self.__repr__()


class ControlAction(IntEnum):
    """
    Specify which action should be executed, e.g. `DISABLE` or `ENABLE`.

    Values
    ------
    - NONE
    - DISABLE
    - ENABLE
    """

    NONE = 0
    DISABLE = auto()
    ENABLE = auto()

    @classmethod
    def _from_raw(cls, en: _ControlAction):
        return ControlAction[en.name]

    def _to_raw(self):
        return _ControlAction[self.name]

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.__repr__()


class URI:
    """
    Class describing a unique resource identifier.

    Parameters
    ----------
    None

    Examples
    --------
    >>> uri = URI("scheme://user:secretpass@example.com:8443/path/to/resource?query=search&sort=asc#section-2")
    """

    def __init__(self, uri: str | Path, scheme: str = "", allow_fragments=True) -> None:
        uri = uri.as_posix() if isinstance(uri, Path) else uri

        self._uri: parse.SplitResult = parse.urlsplit(
            str(uri), scheme=scheme, allow_fragments=allow_fragments
        )

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, URI):
            return parse.urlunsplit(self._uri) == parse.urlunsplit(value._uri)

        return False

    def __ne__(self, value: object, /) -> bool:
        # Needs to be implemented because of __eq__
        return not self.__eq__(value)

    def _to_internal(self):
        return _URI(parse.urlunsplit(self._uri))

    def netloc(self) -> str:
        return self._uri.netloc

    def scheme(self) -> str:
        return self._uri.scheme

    def username(self) -> str | None:
        return self._uri.username

    def password(self) -> str | None:
        return self._uri.password

    def hostname(self) -> str | None:
        return self._uri.hostname

    def port(self) -> int | None:
        return self._uri.port

    def path(self) -> str:
        return self._uri.path

    def query(self) -> str:
        return self._uri.query

    def fragment(self) -> str:
        return self._uri.fragment

    def __repr__(self) -> str:
        return parse.urlunsplit(self._uri)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from pyfdb import ControlAction, ControlIdentifier
from pyfdb._internal import _ControlAction, _ControlIdentifier


def test_enum_control_identifiers():
    assert ControlIdentifier.NONE == 0
    assert ControlIdentifier.LIST == 1
    assert ControlIdentifier.RETRIEVE == 2
    assert ControlIdentifier.ARCHIVE == 4
    assert ControlIdentifier.WIPE == 8
    assert ControlIdentifier.UNIQUEROOT == 16


control_identifiers_enum_variants = [
    (_ControlIdentifier.NONE, ControlIdentifier.NONE),
    (_ControlIdentifier.LIST, ControlIdentifier.LIST),
    (_ControlIdentifier.RETRIEVE, ControlIdentifier.RETRIEVE),
    (_ControlIdentifier.ARCHIVE, ControlIdentifier.ARCHIVE),
    (_ControlIdentifier.WIPE, ControlIdentifier.WIPE),
    (_ControlIdentifier.UNIQUEROOT, ControlIdentifier.UNIQUEROOT),
]


@pytest.mark.parametrize("variant, expected", control_identifiers_enum_variants)
def test_enum_control_identifiers_from_raw(variant, expected):
    assert ControlIdentifier._from_raw(variant) == expected


control_action_enum_variants = [
    (_ControlAction.NONE, ControlAction.NONE),
    (_ControlAction.DISABLE, ControlAction.DISABLE),
    (_ControlAction.ENABLE, ControlAction.ENABLE),
]


@pytest.mark.parametrize("variant, expected", control_action_enum_variants)
def test_enum_control_actions_from_raw(variant, expected):
    assert ControlAction._from_raw(variant) == expected


def test_control_action_mapping():
    assert ControlAction._from_raw(ControlAction.NONE._to_raw()) == ControlAction.NONE
    assert (
        ControlAction._from_raw(ControlAction.DISABLE._to_raw())
        == ControlAction.DISABLE
    )
    assert (
        ControlAction._from_raw(ControlAction.ENABLE._to_raw()) == ControlAction.ENABLE
    )


def test_control_identifier_mapping():
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.NONE._to_raw())
        == ControlIdentifier.NONE
    )
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.LIST._to_raw())
        == ControlIdentifier.LIST
    )
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.RETRIEVE._to_raw())
        == ControlIdentifier.RETRIEVE
    )
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.ARCHIVE._to_raw())
        == ControlIdentifier.ARCHIVE
    )
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.WIPE._to_raw())
        == ControlIdentifier.WIPE
    )
    assert (
        ControlIdentifier._from_raw(ControlIdentifier.UNIQUEROOT._to_raw())
        == ControlIdentifier.UNIQUEROOT
    )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# Cases where we support all selection on the C++ API
#
# list
# status
# wipe
# purge
# stats
# control
# axes

from pathlib import Path

from pyfdb.pyfdb import FDB
from pyfdb.pyfdb_iterator import ControlElement, IndexAxis
from pyfdb.pyfdb_type import ControlAction, ControlIdentifier


def test_list_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    list_iterator = fdb.list({}, level=1)
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 32

    print("----------------------------------")

    list_iterator = fdb.list({}, level=2)
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 32

    print("----------------------------------")

    list_iterator = fdb.list({}, level=3)
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        print(el.uri)
        elements.append(el)

    assert len(elements) == 96


def test_status_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    status_iterator = fdb.status({})

    elements = []

    for el in status_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 32


def test_wipe_dryrun_all(read_write_fdb_setup):
    fdb = FDB(read_write_fdb_setup)
    elements = list(fdb.list({}))
    initial_len = len(elements)

    wipe_iterator = fdb.wipe({})
    wiped_elements = list(wipe_iterator)
    assert len(wiped_elements) > initial_len

    for el in wiped_elements:
        print(el)

    elements_after_wipe = list(fdb.list({}))
    assert initial_len == len(elements_after_wipe)


def test_wipe_all_doit_all(read_write_fdb_setup):
    fdb = FDB(read_write_fdb_setup)

    elements = list(fdb.list({}))
    assert len(elements) > 0

    wipe_iterator = fdb.wipe({}, doit=True)
    wiped_elements = list(wipe_iterator)
    assert len(wiped_elements) > 0

    elements_after_wipe = list(fdb.list({}))
    print(
        f"#Elements before: {len(elements)}, Elements after: {len(elements_after_wipe)}"
    )
    assert len(elements_after_wipe) == 0


def test_purge_all(empty_fdb_setup, build_grib_messages):
    fdb = FDB(empty_fdb_setup)

    initial_elements = list(fdb.list({}))

    assert len(initial_elements) == 0
    print(f"Elements initial: {len(initial_elements)}")

    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    archived_elements = list(fdb.list({}))
    assert len(archived_elements) == 96

    print(f"Elements after archive: {len(archived_elements)}")

    purge_iterator = fdb.purge({}, doit=True)
    purged_output = []

    for el in purge_iterator:
        print(el)
        purged_output.append(el)

    assert len(purged_output) > 0

    elements_after_first_purge = list(fdb.list({}))

    print(f"Elements after purge: {len(elements_after_first_purge)}")
    assert len(archived_elements) == len(elements_after_first_purge)

    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    after_archive_elements = list(fdb.list({}))

    print(
        f"Elements after masking archive: {len(after_archive_elements)}, before: {len(initial_elements)}"
    )
    assert len(after_archive_elements) > len(initial_elements)

    purge_iterator = fdb.purge({})
    purged_output = list(purge_iterator)

    assert len(purged_output) > 0

    elements_after_second_purge = list(fdb.list({}))

    for el in elements_after_second_purge:
        print(el)

    print(f"Elements after second purge: {len(elements_after_second_purge)}")
    assert len(archived_elements) == len(elements_after_second_purge)


def test_stats_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    list_iterator = fdb.stats({})
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 32


def test_control_lock_retrieve_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    print("Retrieve without lock")
    data_handle = fdb.retrieve(
        {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "167/165/166",
            "time": "1800",
        }
    )
    assert data_handle
    data_handle.open()
    assert data_handle.read(4) == b"GRIB"
    data_handle.close()

    assert not (
        read_only_fdb_setup.parent
        / "db_store"
        / "ea:0001:oper:20200101:1800:g"
        / "retrieve.lock"
    ).exists()

    print("Locking entire database for retrieve")

    control_iterator = fdb.control(
        {},
        ControlAction.DISABLE,
        [ControlIdentifier.RETRIEVE],
    )
    assert control_iterator

    elements = []

    el: ControlElement

    for el in control_iterator:
        assert (Path(str(el.location())) / "retrieve.lock").exists()
        elements.append(el)

    assert len(elements) == 32

    # Needed because reuse of databases in the retrieve path is using a cached db
    # object which contains the first (non-locked) status of the db.
    # This can be removed as soon as https://jira.ecmwf.int/browse/FDB-613 is fixed.
    fdb = FDB(read_only_fdb_setup)

    with fdb.retrieve(
        {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "167/165/166",
            "time": "1800",
        },
    ) as data_handle_2:
        assert data_handle_2
        assert data_handle_2.read(4) == b""


def test_index_axis_items_levels_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    index_axis: IndexAxis = fdb.axes({})

    assert len(index_axis) == 11

    print("---------- Level 3: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")

    for k, v in index_axis.items():
        assert index_axis.has_key(k)

    index_axis: IndexAxis = fdb.axes({}, level=2)

    assert len(index_axis) == 8

    print("---------- Level 2: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")

    index_axis: IndexAxis = fdb.axes({}, level=1)

    assert len(index_axis) == 6

    print("---------- Level 1: ----------")

    for k, v in index_axis.items():
        print(f"k={k} | v={v}")



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import json
from collections.abc import Collection
from pathlib import Path
from typing import Any, Optional

import yaml

from pyfdb_bindings import pyfdb_bindings as pyfdb_internal

# Initial setup of binding via eckit main
pyfdb_internal.init_bindings()

InternalMarsSelection = dict[str, Collection[str]]
"""
This is the internal representation of a MARS selection

This is a key-value map, mapping MARS keys to a string resembling values, value lists or value ranges.
"""

InternalMarsIdentifier = list[tuple[str, str]]
"""
This is the internal representation of a MARS identifier

This is a key-value list, mapping MARS keys to a string resembling a singluar value, see https://github.com/ecmwf/datacube-spec.
"""


class FDBToolRequest:
    """
    FDBToolRequest object

    Parameters
    ----------
    `key_values` : `MarsSelection` | None, *optional*, default: `None`
        Dictionary of key-value pairs which describe the MARS selection
    `all` : `bool`, *optional*, default: `False`
        Should a tool request be interpreted as querying all data? True if so, False otherwise.
    `minimum_key_set` : `list[str]`, *optional*, default: `None`
        Define the minimum set of keys that must be specified. This prevents inadvertently exploring the entire FDB.

    Returns
    -------
    FDBToolRequest object

    Examples
    --------
    >>> request = FDBToolRequest(
    ...     {
    ...         "class": "ea",
    ...         "domain": "g",
    ...         "expver": "0001",
    ...         "stream": "oper",
    ...         "date": "20200101",
    ...         "time": "1800",
    ...     },
    ...     # all = False,
    ...     # minimum_key_set = None,
    ... )
    """

    def __init__(
        self,
        key_values=None | InternalMarsSelection,
        all: bool = False,
        minimum_key_set: list[str] | None = None,
    ) -> None:
        if key_values is not None:
            key_values = key_values

        if minimum_key_set is None:
            minimum_key_set = []

        self.tool_request = pyfdb_internal.FDBToolRequest(
            key_values, all, minimum_key_set
        )

    @classmethod
    def from_internal_mars_selection(
        cls, selection: InternalMarsSelection | None
    ) -> "FDBToolRequest":
        if selection is None or len(selection) == 0:
            return cls(key_values={}, all=True)

        for k, v in selection.items():
            if isinstance(v, str):
                raise ValueError(f"Expecting collection of values for key {k}, not {v}")

        return cls(key_values=selection, all=False)

    def __repr__(self) -> str:
        return repr(self.tool_request)


class ConfigMapper:
    @classmethod
    def to_json(cls, config: str | dict | Path | None) -> Optional[str]:
        """
        Mapper for creating a configuration string from given config

        Parameters
        ----------
        `config`: `str` | `dict` | `Path`
                System or user configuration

        Note
        ----
        *In case you want to supply a config to a FDB, hand in the needed arguments to the FDB directly.*
        For usage examples see the comments of FDB constructor.

        The config parameter can have different types, depending on the meaning:
            - `str` is used as a yaml representation to parse the config
            - `dict` is interpreted as hierarchical format to represent a config, see example
            - `Path` is interpreted as a location of the config and read as a YAML file

        Returns
        -------
        JSON string representing the given FDB configuration.

        Examples
        --------
        >>> config = ConfigMapper.to_json(
        ...     {
        ...         "type":"local",
        ...         "engine":"toc",
        ...         "schema": "<schema_path>",
        ...         "spaces":[
        ...             {
        ...                 "handler":"Default",
        ...                 "roots":[
        ...                     {"path": "<db_store_path>"},
        ...                 ],
        ...             }
        ...         ],
        ...     })
        """

        if config is None:
            return config

        config_result = None

        if isinstance(config, Path):
            config_result = yaml.safe_load(config.read_text())
        elif isinstance(config, dict):
            config_result = config
        elif isinstance(config, str):
            config_result = yaml.safe_load(config)
        else:
            raise ValueError("Config: Unknown config type, must be str, dict or Path.")

        return json.dumps(config_result)

    @classmethod
    def from_json(cls, config: str) -> dict[str, Any]:
        return json.loads(config)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.


from pyfdb.pyfdb_type import URI, DataHandle, ControlAction, ControlIdentifier, MarsSelection
from pyfdb.pyfdb import FDB

from pyfdb.pyfdb_iterator import (
    ListElement,
    StatusElement,
    WipeElement,
    PurgeElement,
    StatsElement,
    ControlElement,
    IndexAxis,
)


__all__ = [
    "FDB",
    "URI",
    "DataHandle",
    "ListElement",
    "WipeElement",
    "StatusElement",
    "ControlAction",
    "ControlIdentifier",
    "PurgeElement",
    "StatsElement",
    "ControlElement",
    "IndexAxis",
    "MarsSelection",
]



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import logging
from collections.abc import Collection, ItemsView, Iterator, KeysView, ValuesView
from enum import IntEnum, auto
from typing import Optional

from pyfdb._internal import (
    ControlElement as _ControlElement,
)
from pyfdb._internal import (
    IndexAxis as _IndexAxis,
)
from pyfdb._internal import (
    ListElement as _ListElement,
)
from pyfdb._internal import (
    PurgeElement as _PurgeElement,
)
from pyfdb._internal import (
    StatsElement as _StatsElement,
)
from pyfdb._internal import (
    WipeElement as _WipeElement,
)
from pyfdb._internal import (
    WipeElementType as _WipeElementType,
)
from pyfdb._internal.pyfdb_internal import InternalMarsSelection
from pyfdb.pyfdb_type import URI, ControlIdentifier, DataHandle, MarsSelection, UserInputMapper

logger = logging.getLogger(__name__)


class ListElement:
    """Element returned from a listing command"""

    def __init__(self, list_element: _ListElement, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a ListElement from user code is not supported.")
        self._element: _ListElement = list_element

    def has_location(self) -> bool:
        """
        Does the `ListElement` have a location

        Returns
        -------
        `bool`
            `True` if this element has a location, `False` otherwise.

        Note
        ----
        Only for `ListElement`s which are on the third level of the schema.
        """
        return self._element.has_location()

    def offset(self) -> Optional[int]:
        """
        Offset within the file associated with the `ListElement`

        Returns
        -------
        `Optional[int]`
            Offset in bytes in the data file of the FDB, if `ListElement` is on `level` 3, None otherwise.

        Note
        ----
        Only for `ListElement`s which are on the third level of the schema.
        """
        if self.has_location():
            return self._element.offset()
        logger.info("Asking for offset on list element without location. Did you specify `level=3` in a list call?")
        return None

    def length(self) -> Optional[int]:
        """
        Size of the `ListElement` within the file associated.

        Returns
        -------
        `Optional[int]`
            Size in bytes in the data file of the FDB, if `ListElement` is on `level` 3, None otherwise.

        Note
        ----
        Only for `ListElement`s which are on the third level of the schema.
        """
        if self.has_location():
            return self._element.length()
        logger.info("Asking for length on list element without location. Did you specify `level=3` in a list call?")
        return None

    def combined_key(self) -> dict[str, str]:
        """
        Return the MARS keys of the `ListElement`

        Returns
        -------
        `dict[str, str]`
            Dictionary containing all metadata for the `ListElement`

        Note
        ----
        Depending on the `level` specified in the `list` command, the returned dictionary contains
        all available keys from schema levels up to and including the requested level; keys that
        exist only at deeper levels are omitted.

        Examples
        --------
        >>> list_iterator = fdb.list(selection, level=3)
        >>> assert list_iterator

        >>> elements = list(list_iterator)

        >>> for element in elements:
        >>>     print(element.combined_key())

        Output:

        ``
        ...
        { 'class': 'ea', 'date': '20200104', ... , 'type': 'an' }
        ...
        ``
        """
        return self._element.combined_key()

    def keys(self) -> list[dict[str, str]]:
        """
        Return the MARS keys of the `ListElement` separated by their level in the schema.

        Returns
        -------
        `list[dict[str, str]]`
            List containing MARS keys and their values for the `ListElement`. The keys are split
            by their level in the schema

        Note
        ----
        Depending on the `level` specified in the `list` command, the returned dictionary contains
        all available keys from schema levels up to and including the requested level; keys that
        exist only at deeper levels are omitted.

        Examples
        --------
        >>> list_iterator = fdb.list(selection, level=3)
        >>> assert list_iterator

        >>> elements = list(list_iterator)

        >>> for element in elements:
        >>>     print(element.keys())

        Output:

        ``
        ...
        [{'class': 'ea', ... , 'time': '2100'}, {'levtype': 'sfc', 'type': 'an'}, {'param': '167', 'step': '0'}]
        ...
        ``
        """
        return self._element.keys()

    @property
    def data_handle(self) -> Optional[DataHandle]:
        """
        Access the `DataHandle`

        Returns
        -------
        `DataHandle`
            Data Handle for accessing the data of the list element

        Examples
        --------
        >>> data_handle = list_element.data_handle
        >>> if data_handle is not None:
        >>>     data_handle.open()
        >>>     data_handle.read(4)
        >>>     data_handle.close()

        Output:

        ``b"GRIB"``
        """
        if self.has_location():
            return DataHandle(self._element.data_handle(), _internal=True)

        logger.info(
            "Asking for data handle on list element without location. Did you specify `level=3` in a list call?"
        )

        return None

    @property
    def uri(self) -> Optional[URI]:
        """
        Access the URI of the list element

        Returns
        -------
        `URI`
            URI of the data

        Examples
        --------
        >>> uri = list_element.uri
        >>> print(uri)

        Output:

        ``<path/to/data_file>``
        """
        if self.has_location():
            return URI(self._element.uri())

        logger.info(
            "Asking for data handle on list element without location. Did you specify `level=3` in a list call?"
        )

        return None

    def __repr__(self) -> str:
        return str(self._element)


class WipeElementType(IntEnum):
    """

    Values
    ------
    - ERROR
    - CATALOGUE_INFO
    - CATALOGUE
    - CATALOGUE_INDEX
    - CATALOGUE_SAFE
    - CATALOGUE_CONTROL
    - STORE
    - STORE_AUX
    - STORE_SAFE
    - UNKNOWN
    """

    ERROR = 0
    CATALOGUE_INFO = auto()
    CATALOGUE = auto()
    CATALOGUE_INDEX = auto()
    CATALOGUE_SAFE = auto()
    CATALOGUE_CONTROL = auto()
    STORE = auto()
    STORE_AUX = auto()
    STORE_SAFE = auto()
    UNKNOWN = auto()

    @classmethod
    def _from_raw(cls, en: _WipeElementType):
        return WipeElementType[en.name]

    def _to_raw(self):
        return _WipeElementType[self.name]

    def __repr__(self) -> str:
        return self.name

    def __str__(self) -> str:
        return self.__repr__()


class WipeElement:
    def __init__(self, wipe_element: str, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a WipeElement from user code is not supported.")
        self._element: _WipeElement = wipe_element

    def type(self) -> WipeElementType:
        return self._element.type()

    def msg(self) -> str:
        return self._element.msg()

    def uris(self) -> list[URI]:
        return [URI(uri) for uri in self._element.uris()]

    def __repr__(self) -> str:
        return str(self._element)


class StatusElement:
    def __init__(self, control_element: _ControlElement, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a StatusElement from user code is not supported.")
        self.element: _ControlElement = control_element

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, (StatusElement, ControlElement)):
            return (
                self.location() == other.location()
                and self.controlIdentifiers() == other.controlIdentifiers()
                and self.key() == other.key()
            )

        return False

    def __ne__(self, value: object, /) -> bool:
        # Needs to be implemented because of __eq__
        return not self.__eq__(value)

    def location(self) -> URI:
        return URI(self.element.location())

    def controlIdentifiers(self) -> list[ControlIdentifier]:
        return [ControlIdentifier._from_raw(el) for el in self.element.controlIdentifiers()]

    def key(self) -> MarsSelection:
        return self.element.key()

    def __repr__(self) -> str:
        return f"StatusElement(control_identifiers={self.controlIdentifiers()}, key={self.key()}, location={self.location()})"


class PurgeElement:
    def __init__(self, purge_element: str, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a PurgeElement from user code is not supported.")
        self.element: _PurgeElement = purge_element

    def __repr__(self) -> str:
        return str(self.element)


class StatsElement:
    def __init__(self, stats_element: _StatsElement, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a StatsElement from user code is not supported.")
        self.element: _StatsElement = stats_element

    def db_statistics(self) -> str:
        return self.element.db_statistics()

    def index_statistics(self) -> str:
        return self.element.index_statistics()

    def __repr__(self) -> str:
        return str(self.element)


class ControlElement:
    def __init__(self, control_element: _ControlElement, *, _internal=False) -> None:
        if not _internal:
            raise TypeError("Creating a ControlElement from user code is not supported.")
        self.element: _ControlElement = control_element

    def __eq__(self, other: object, /) -> bool:
        if isinstance(other, (StatusElement, ControlElement)):
            return (
                self.location() == other.location()
                and self.controlIdentifiers() == other.controlIdentifiers()
                and self.key() == other.key()
            )

        return False

    def __ne__(self, value: object, /) -> bool:
        # Needs to be implemented because of __eq__
        return not self.__eq__(value)

    def location(self) -> URI:
        return URI(self.element.location())

    def controlIdentifiers(self) -> list[ControlIdentifier]:
        return [ControlIdentifier._from_raw(el) for el in self.element.controlIdentifiers()]

    def key(self) -> MarsSelection:
        return self.element.key()

    def __repr__(self) -> str:
        return f"ControlElement(control_identifiers={self.controlIdentifiers()}, key={self.key()}, location={self.location()})"


class IndexAxis(InternalMarsSelection):
    """
    `IndexAxis` class representing axes and their extent. The class implements all Dictionary
    functionalities. Key are the corresponding FDB keys (axes) and values are the values defining the extent
    of an axis.

    """

    def __init__(self, index_axis: _IndexAxis | MarsSelection) -> None:
        if isinstance(index_axis, _IndexAxis):
            self.index_axis: _IndexAxis = index_axis
        elif isinstance(index_axis, dict):
            internal_mars_selection = UserInputMapper.map_selection_to_internal(index_axis)
            self.index_axis: _IndexAxis = _IndexAxis(internal_mars_selection)
        else:
            raise TypeError(
                f"IndexAxis: Unknown type {type(index_axis)} for creating IndexAxis. Only `IndexAxis` or `dict[str, Collection[str]]` are allowed."
            )

    def __repr__(self) -> str:
        return repr(self.index_axis)

    def __getitem__(self, key: str) -> Collection[str]:
        return self.index_axis[key]

    def __setitem__(self, key, value):
        raise TypeError("This mapping is read-only")

    def __delitem__(self, key):
        raise TypeError("This mapping is read-only")

    def clear(self):
        raise TypeError("This mapping is read-only")

    def __len__(self) -> int:
        return len(self.index_axis)

    def __iter__(self) -> Iterator[str]:
        return iter(self.index_axis.keys())

    def __eq__(self, value) -> bool:
        dict_list_values = {}

        if isinstance(value, (dict, IndexAxis)):
            for k, v in value.items():
                if isinstance(v, str):
                    dict_list_values[k] = [v]
                elif not isinstance(v, Collection):
                    dict_list_values[k] = [str(v)]
                else:
                    dict_list_values[k] = v

            return dict(self.items()) == dict(dict_list_values.items())

        return False

    def __ne__(self, value: object, /) -> bool:
        # Needs to be implemented because of __eq__
        return not self.__eq__(value)

    def has_key(self, k) -> bool:
        return k in self.index_axis

    def keys(self) -> KeysView[str]:
        return KeysView(self)

    def values(self) -> ValuesView[Collection[str]]:
        return ValuesView(self)

    def items(self) -> ItemsView[str, Collection[str]]:
        return ItemsView(self)

    def __contains__(self, item: object):
        return item in self.index_axis



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
z3FDB enables to create views into FDB where the view is a Zarr array.

Views are defined by one or more MARS requests. Each keyword in the MARS
request with more than one value defines an 'Axis'. 'Axis' from MARS requests
need to be mapped to 'Axis' in the Zarr array. This mapping can be a 1-1 or
many-1 mapping, allowing to create a time based axis in the Zarr array that
is composed from the 'date' and 'time' keyword when dealing with climate data.

For example, the request
``"..., date=1970-01-1/to/2020-12-31, time=00/06/12/18, ..."`` spans two axis
'date' and 'time'. If you want to work on a unified time axis, then you  can
use the following :class:`AxisDefinition` to map accordingly:

Example::

    AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE)

This defines an 'Axis' in the Zarr array that follows 'date' and 'time' from
the MARS request, where the rightmost references 'Axis' ('time') is varying
fastest.

You can combine multiple MARS request into one view. This is useful if you
want to access surface and pressure level data in one view. In this case you
need to select on which 'Axis' of the Zarr array the requests extend each
other. The remaining axis have to have the same cardinality.

.. code-block:: python

   builder.add_part(
       "type=an,"
       "class=ea,"
       "domain=g,"
       "expver=0001,"
       "stream=oper,"
       "date=2020-01-01/2020-01-02,"
       "levtype=sfc,"
       "step=0,"
       "param=165/166,"
       "time=0/to/21/by/3",
       [
           AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
           AxisDefinition(["param"], Chunking.SINGLE_VALUE)
       ],
       ExtractorType.GRIB,
   )
   builder.add_part(
       "type=an,"
       "class=ea,"
       "domain=g,"
       "expver=0001,"
       "stream=oper,"
       "date=2020-01-01/2020-01-02,"
       "levtype=pl,"
       "step=0,"
       "param=131/132,"
       "levelist=50/100,"
       "time=0/to/21/by/3",
       [
           AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
           AxisDefinition(["param", "levelist"], Chunking.SINGLE_VALUE)
       ],
       ExtractorType.GRIB,
   )
   builder.extend_on_axis(1)
   store = builder.build()

The created Zarr array will always have the actual data points available as
the final 'Axis'.

.. code-block:: text

    arr[0][0][0][0]
        ^  ^  ^  ^
        |  |  | Index in field  -> Implicit
        |  | Ensemble           -> Created from an AxisDefinition
        | Step                  -> Created from an AxisDefinition
       Date                     -> Created from an AxisDefinition

"""

from pychunked_data_view.chunked_data_view import (
    AxisDefinition,
    Chunking,
    ExtractorType,
)
from z3fdb.simple_store_builder import (
    SimpleStoreBuilder,
)

from z3fdb.z3fdb_error import Z3fdbError

__all__ = [
    "AxisDefinition",
    "Chunking",
    "ExtractorType",
    "SimpleStoreBuilder",
    "Z3fdbError",
]



# (C) Copyright 2011- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from findlibs import Path
import pyfdb
import pytest
import pathlib

import itertools
import shutil
import yaml
import git

import eccodes as ec

from doctest import ELLIPSIS

from sybil import Sybil
from sybil.parsers.rest import DocTestParser, PythonCodeBlockParser


def sybil_setup(namespace):
    # there are better ways to do temp directories, but it's a simple example:
    namespace["pyfdb"] = pyfdb


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(optionflags=ELLIPSIS),
        PythonCodeBlockParser(),
    ],
    patterns=["*.rst", "*.py"],
    fixtures=["fdb_config_path", "data_path", "test_case_tmp"],
    setup=sybil_setup,
).pytest()


def get_git_root(path) -> Path:
    git_repo = git.Repo(path, search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")
    return Path(git_root)


@pytest.fixture(scope="function")
def data_path(doctest_namespace) -> pathlib.Path:
    """
    Provides path to test data
    """
    path = get_git_root(__file__).resolve() / "tests" / "pyfdb" / "data"
    assert path.exists()
    doctest_namespace["data_path"] = path
    return path


@pytest.fixture(scope="function")
def test_data_path() -> pathlib.Path:
    """
    Provides path to test data
    """

    git_repo = git.Repo(__file__, search_parent_directories=True)
    git_root = pathlib.Path(git_repo.git.rev_parse("--show-toplevel"))
    path = git_root / "tests" / "pyfdb" / "data"
    assert path.exists()
    return path


@pytest.fixture(scope="function")
def test_case_tmp(tmp_path_factory) -> pathlib.Path:
    return tmp_path_factory.mktemp("session_data")


@pytest.fixture(scope="function")
def build_grib_messages(data_path, test_case_tmp) -> pathlib.Path:
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")
    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()
    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing datapoints
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")
    ec.codes_set_string(gid, "levtype", "sfc")
    ec.codes_set_values(gid, list(range(0, count_values)))

    dates = [20200101, 20200102, 20200103, 20200104]
    times = [0, 300, 600, 900, 1200, 1500, 1800, 2100]
    parameters = [167, 131, 132]

    messages = test_case_tmp / "test_data.grib"
    with open(messages, "wb") as out:
        for date, time, parameter in itertools.product(dates, times, parameters):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages


@pytest.fixture(scope="function", autouse=True)
def fdb_config_path(
    data_path, test_case_tmp, build_grib_messages, doctest_namespace
) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    Test FDB currently reads all grib files in `tests/data`
    This setup can be shared between tests as we will only read
    data from this FDB
    """
    schema_path_src = data_path / "schema"
    assert schema_path_src.is_file()
    schema_path = test_case_tmp / "schema"
    shutil.copy(schema_path_src, schema_path)

    db_store_path = test_case_tmp / "db_store"
    db_store_path.mkdir()
    fdb_config = {
        "type": "local",
        "engine": "toc",
        "schema": str(schema_path),
        "spaces": [
            {
                "handler": "Default",
                "roots": [
                    {"path": str(db_store_path)},
                ],
            }
        ],
    }
    fdb_config_str = yaml.dump(fdb_config)
    fdb_config_path = test_case_tmp / "fdb_config.yaml"
    fdb_config_path.write_text(fdb_config_str)
    fdb = pyfdb.FDB(fdb_config_str)
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()
    doctest_namespace["fdb_config_path"] = fdb_config_path
    return fdb_config_path



import logging

import zarr

from z3fdb import (
    AxisDefinition,
    Chunking,
    ExtractorType,
    SimpleStoreBuilder,
)

logging.basicConfig(level=logging.DEBUG)


def test_random_axis_retrieval(read_only_fdb_pattern_setup) -> None:
    """Test the retrieval of random axis order in a MARS request

    If we permute the order of the request and the values for each individual key,
    we expect this to be part of the resulting data view, as well.

    This means that the permutation needs to be tested and checked against. The values
    for the 'fields' in the data view are injected by the fixture in conftest.py
    """
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166,"
        "time=18/0/12/6",
        [
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern

    def value_at(param, time, date):
        date_perm = [2, 0, 1]
        param_perm = [2, 0, 1]
        time_perm = [3, 0, 2, 1]
        return (
            param_perm[param] * 1
            + time_perm[time] * 3 * 1
            + date_perm[date] * 4 * 3 * 1
        )

    # This is the first entire in the fixture which creates constant fields
    # therefore the field has to be a constant 0
    assert value_at(1, 1, 1) == 0

    # Same but for the last entry, therefore the field has to be
    # a constant 35 = 3 dates a 4 time points a 3 params a 1 step - 1 (because 0 indexed)
    assert value_at(0, 0, 0) == 35

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                assert all(data[date, time, param, 0] == value_at(param, time, date))


def test_random_axis_retrieval_swapped_axis_and_request(
    read_only_fdb_pattern_setup,
) -> None:
    """Test the retrieval of random axis order in a MARS request

    If we permute the order of the request and the values for each individual key,
    we expect this to be part of the resulting data view, as well. In this case the AxisDefinitions
    are permuted, as well and we check for the correct behavior

    This means that the permutation needs to be tested and checked against. The values
    for the 'fields' in the data view are injected by the fixture in conftest.py
    """
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "time=18/0/12/6,"
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166",
        [
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern

    def value_at(param, time, date):
        date_perm = [2, 0, 1]  # Due to switching order of date in the request
        param_perm = [2, 0, 1]
        time_perm = [3, 0, 2, 1]
        return (
            param_perm[param] * 1
            + time_perm[time] * 3 * 1
            + date_perm[date] * 4 * 3 * 1
        )

    # This is the first entire in the fixture which creates constant fields
    # therefore the field has to be a constant 0
    assert value_at(1, 1, 1) == 0

    # Same but for the last entry, therefore the field has to be
    # a constant 35 = 3 dates a 4 time points a 3 params a 1 step - 1 (because 0 indexed)
    assert value_at(0, 0, 0) == 35

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                assert all(data[time, 0, param, date] == value_at(param, time, date))


def test_random_axis_retrieval_swapped_axis_and_request_combined_axis(
    read_only_fdb_pattern_setup,
) -> None:
    """Test the retrieval of random axis order in a MARS request

    If we permute the order of the request and the values for each individual key,
    we expect this to be part of the resulting data view, as well. In this case the AxisDefinitions
    are permuted and combined, as well. Additionally the chunking is disabled which leads to the whole
    request being fetch at once.

    This means that the permutation needs to be tested and checked against. The values
    for the 'fields' in the data view are injected by the fixture in conftest.py
    """
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)

    # The individual values are scrambled to check for persistent retrieval on the FDB side
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "time=18/0/12/6,"
        "date=2020-01-03/2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=167/165/166",
        [
            AxisDefinition(["time", "step", "param", "date"], Chunking.NONE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern

    def value_at(param, time, date):
        date_perm = [2, 0, 1]  # Due to switching order of date in the request
        param_perm = [2, 0, 1]
        time_perm = [3, 0, 2, 1]
        return (
            param_perm[param] * 1
            + time_perm[time] * 3 * 1
            + date_perm[date] * 4 * 3 * 1
        )

    def linearize(time, param, date):
        return date + param * 3 + time * 3 * 3

    # This is the first entire in the fixture which creates constant fields
    # therefore the field has to be a constant 0
    assert value_at(1, 1, 1) == 0

    # Same but for the last entry, therefore the field has to be
    # a constant 35 = 3 dates a 4 time points a 3 params a 1 step - 1 (because 0 indexed)
    assert value_at(0, 0, 0) == 35

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                assert all(
                    data[linearize(time, param, date)] == value_at(param, time, date)
                )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

"""
This module contains all code internal to z3fdb.

This is not part of the supported interface.
"""

import json

try:
    from collections.abc import Buffer
except ImportError:
    # Python < 3.12
    from typing import Union

    Buffer = Union[bytes, bytearray, memoryview]

from typing import AsyncIterator, Iterable

import numpy as np
import itertools

from zarr.abc import store
from zarr.core.buffer import default_buffer_prototype
from zarr.core.buffer.core import Buffer as AbstractBuffer
from zarr.core.buffer.core import BufferPrototype
from zarr.core.buffer.cpu import Buffer as CpuBuffer
from zarr.core.common import BytesLike

from functools import cache
from typing import Self

from pychunked_data_view import (
    ChunkedDataView,
)

from z3fdb.z3fdb_error import Z3fdbError
from dataclasses import KW_ONLY, asdict, dataclass, field
from typing import Any, Optional, Sequence


def to_cpu_buffer(d: dict) -> CpuBuffer:
    return CpuBuffer.from_bytes(json.dumps(d).encode("utf-8"))


def from_cpu_buffer(buf: CpuBuffer) -> dict:
    return json.loads(buf.to_bytes().decode("utf-8"))


@dataclass(frozen=True)
class DotZarrAttributes:
    _: KW_ONLY
    copyright: str = "ecmwf"
    zarr_format: int = 3
    variables: Sequence[dict] = field(default_factory=list)

    def asbytes(self) -> bytes:
        return json.dumps(asdict(self)).encode("utf8")

    def asstring(self) -> str:
        return json.dumps(asdict(self), indent=2)


@dataclass(frozen=True)
class MetadataConfiguration:
    _: KW_ONLY
    name: str
    configuration: dict[str, Any]


@dataclass(frozen=True)
class ChunkGridMetadata(MetadataConfiguration):
    def __init__(self, chunks) -> None:
        super().__init__(name="regular", configuration={"chunk_shape": chunks})


@dataclass
class DotZarrArrayJson:
    """
    Generates the .zarr metadata for an array.

    If additional fields are introduced, read the documentation about must_understand
    https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#id13

    Most of what happens here has been reverse-engineered from the zarr-python code.
    """

    _: KW_ONLY
    zarr_format: int = 3
    node_type: str = "array"
    shape: tuple[int, ...]
    data_type: str | Sequence[str] | MetadataConfiguration
    chunk_grid: MetadataConfiguration
    chunk_key_encoding: MetadataConfiguration = MetadataConfiguration(
        name="default", configuration={"separator": "/"}
    )
    codecs: Sequence[MetadataConfiguration] = field(
        default_factory=lambda: [
            MetadataConfiguration(name="bytes", configuration={"endian": "little"})
        ]
    )
    fill_value: bool | int | float | None = None
    attributes: Optional[dict[str, str] | DotZarrAttributes] = field(
        default=DotZarrAttributes()
    )
    storage_transformers: Optional[Sequence[MetadataConfiguration]] = None
    dimension_names: Optional[Sequence[str | None]] = None
    # INFO: If additional fields are introduced, read the documentation about must_understand
    # https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#id13


@dataclass()
class DotZarrGroupJson:
    """
    Generates the .zarr metadata for a group.

    If additional fields are introduced, read the documentation about must_understand
    https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#id13

    Most of what happens here has been reverse-engineered from the zarr-python code.
    """

    _: KW_ONLY
    zarr_format: int = 3
    node_type: str = "group"
    attributes: Optional[dict[str, str] | DotZarrAttributes] = DotZarrAttributes()
    # INFO: If additional fields are introduced, read the documentation about must_understand
    # https://zarr-specs.readthedocs.io/en/latest/v3/core/v3.0.html#id13


class FdbSource:
    """
    Uses FDB as a backend.
    Data is retrieved from FDB and assembled on each access.
    """

    def __init__(
        self,
        chunked_data_view: ChunkedDataView,
    ) -> None:
        self._chunked_data_view = chunked_data_view

        self._shape = self._chunked_data_view.shape()
        self._chunks = self._chunked_data_view.chunkShape()
        self._chunks_per_dimension = self._chunked_data_view.chunks()

    def create_dot_zarr_json(self) -> CpuBuffer:
        return to_cpu_buffer(
            asdict(
                DotZarrArrayJson(
                    shape=self._shape,
                    chunk_grid=ChunkGridMetadata(chunks=self._chunks),
                    data_type="float32",
                    fill_value=-1.0,
                )
            )
        )

    def __contains__(self, key: tuple[int, ...]) -> bool:
        if len(key) != len(self._shape):
            return False
        if any(
            k < 0 or k >= limit for k, limit in zip(key, self._chunks_per_dimension)
        ):
            return False
        return True

    def chunks(self) -> tuple[int, ...]:
        return self._chunks_per_dimension

    def __getitem__(self, key: tuple[int, ...]) -> CpuBuffer:
        if len(key) != len(self._shape):
            raise KeyError
        if any(
            k < 0 or k >= limit for k, limit in zip(key, self._chunks_per_dimension)
        ):
            raise KeyError
        return CpuBuffer.from_bytes(self._chunked_data_view.at(key))


class FdbZarrArray:
    def __init__(self, *, name: str = "", datasource: FdbSource):
        self._name = name
        self._datasource = datasource
        self._metadata = self._datasource.create_dot_zarr_json()

    def __getitem__(self, key: tuple[str, ...]) -> Buffer | None:
        if key[0] == "zarr.json":
            return self._metadata
        if len(key) > 1:
            assert key[0] == "c"  # Zarr v3 for chunks
            chunk_ids = (int(c) for c in key[1:])
            return self._datasource[*chunk_ids]

    @property
    def name(self) -> str:
        return self._name

    @cache
    def paths(self) -> list[str]:
        """
        Zarr paths associated to this array, this includes .zarray, .zattrs and all chunks.

        Returns
        -------
        list[str]
            A list of paths belonging to this group
        """
        files = ["zarr.json"]
        if len(chunks_per_axis := self._datasource.chunks()) > 0:
            tuples = itertools.product(*[np.arange(0, x) for x in chunks_per_axis])
            chunk_names = ["/".join([str(i) for i in ["c", *t]]) for t in tuples]
            files += chunk_names
        return files

    async def list_prefix(self, prefix: str) -> AsyncIterator[str]:
        new_prefix = prefix.removeprefix("/" + self._name)
        new_prefix_token = new_prefix.split("/")

        if len(new_prefix_token) == 1:
            yield self.name


class FdbZarrGroup:
    def __init__(
        self,
        *,
        name: str = "",
        children: list[Self | FdbZarrArray] = [],
    ):
        self._name = name
        self._metadata = to_cpu_buffer(asdict(DotZarrGroupJson()))
        self._attributes = to_cpu_buffer(asdict(DotZarrAttributes()))
        for c in children:
            if c.name == "":
                raise Z3fdbError(
                    "A group with the empty name can only be the root group."
                )
        self._children = {c.name: c for c in children}

    def __getitem__(self, key: tuple[str, ...]) -> Buffer | None:
        if len(key) == 1:
            if key[0] == "zarr.json":
                return self._metadata
        else:
            return self._children[key[0]][*key[1:]]
        raise KeyError(f"Unknown key {key}")

    @property
    def name(self) -> str:
        return self._name

    @property
    def children(self) -> list["FdbZarrArray | FdbZarrGroup"]:
        return list(self._children.values())

    def paths(self) -> list[str]:
        """
        Zarr paths associated to this group, excluding child groups or arrays.

        Returns
        -------
        list[str]
            A list of paths belonging to this group
        """
        return ["zarr.json"]

    async def list_prefix(self, prefix: str) -> AsyncIterator[str]:
        new_prefix = prefix.removeprefix("/" + self._name)
        new_prefix_token = new_prefix.split("/")

        assert len(new_prefix_token) > 0

        if len(new_prefix_token) == 1:
            for k, _ in self._children.items():
                yield k
        else:
            async for i in self._children[new_prefix_token[0]].list_prefix(new_prefix):
                yield i


class FdbZarrStore(store.Store):
    """Provide access to FDB with a MutableMapping.

    .. note:: This is an experimental feature.

    Requires the `pyFDB <https://redis-py.readthedocs.io/>`_
    package to be installed.

    Parameters
    ----------
    """

    def __init__(self, child: FdbZarrGroup | FdbZarrArray):
        super().__init__(read_only=True)

        self._child = child
        self._known_paths = self._build_paths(self._child)
        self._zmetadata = self._consolidate()

    def _build_paths(self, item, parent_path=None) -> list[str]:
        path = f"{parent_path}/{item.name}" if parent_path else item.name
        files = [f"{path}/{f}" if path != "" else f for f in item.paths()]

        if isinstance(item, FdbZarrGroup):
            for child in item.children:
                files += self._build_paths(child, path)

        return files

    def _consolidate(self):
        consolidated_metatdata = {"metadata": {}}

        ENDINGS = ["zarr.json"]

        filtered_paths = (
            path
            for path in self._known_paths
            for key in ENDINGS
            if path is not None and path.endswith(key)
        )

        for path in filtered_paths:
            keys = path.split("/")
            info = self._child[*keys]

            if info is None:
                continue

            consolidated_metatdata["metadata"].update(
                {path: json.loads(info.to_bytes())}
            )

        # Create Buffer
        return CpuBuffer.from_bytes(json.dumps(consolidated_metatdata).encode("utf-8"))

    async def __getitem__(self, key) -> AbstractBuffer | None:
        if key == ".zmetadata":
            return self._zmetadata
        if key == "zarr.json":
            return self._child._metadata

        keys = key.split("/")
        return self._child[*keys]

    def __iter__(self):
        yield from iter(self._known_paths)

    def __len__(self):
        return len(self._known_paths)

    def __setitem__(self, _k, _v):
        raise Z3fdbError("Views into FDB are not writable")

    def __delitem__(self, _k):
        raise Z3fdbError("Views into FDB are not writable")

    def __contains__(self, key) -> bool:
        return key in self._known_paths

    def __eq__(self, value: object) -> bool:
        return self._child == value._child and self._known_paths == value._known_paths

    async def get(
        self,
        key: str,
        prototype: BufferPrototype = default_buffer_prototype(),
        byte_range: store.ByteRequest | None = None,
    ) -> Buffer | None:
        return await self.__getitem__(key)

    def get_partial_values(
        self,
        prototype: BufferPrototype,
        key_ranges: Iterable[tuple[str, store.ByteRequest | None]],
    ) -> list[Buffer | None]:
        pass

    async def exists(self, key: str) -> bool:
        return key in self._known_paths

    @property
    def supports_writes(self) -> bool:
        return False

    async def set(self, key: str, value: AbstractBuffer) -> None:
        raise Z3fdbError("Views into FDB are not writable")

    async def set_if_not_exists(self, key: str, value: AbstractBuffer) -> None:
        raise Z3fdbError("Views into FDB are not writable")

    async def _set_many(self, values: Iterable[tuple[str, AbstractBuffer]]) -> None:
        return await super()._set_many(values)

    @property
    def supports_deletes(self) -> bool:
        return False

    async def delete(self, key: str) -> None:
        raise Z3fdbError("Views into FDB are not writable")

    @property
    def supports_partial_writes(self) -> bool:
        return False

    async def set_partial_values(
        self, key_start_values: Iterable[tuple[str, int, BytesLike]]
    ) -> None:
        raise Z3fdbError("Views into FDB are not writable")

    @property
    def supports_listing(self) -> bool:
        return True

    async def list(self) -> AsyncIterator[str]:
        for i in self._known_paths:
            yield i

    def list_prefix(self, prefix: str) -> AsyncIterator[str]:
        return self._child.list_prefix(prefix)

    def list_dir(self, prefix: str) -> AsyncIterator[str]:
        return self._build_paths(self._child, parent_path=prefix)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import io
import shutil
import tempfile

import pytest

from pyfdb.pyfdb import FDB


def test_datahandle_repr(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)

    assert "Closed" in repr(data_handle)
    data_handle.open()
    assert "Opened" in repr(data_handle)


def test_datahandle_not_opened_before_read(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)

    assert data_handle

    with pytest.raises(
        RuntimeError,
        match="DataHandle: Read occured before the handle was opened. Must be opened first.",
    ):
        assert data_handle.read(4) == b"GRIB"


def test_datahandle_consecutive_read(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)

    assert data_handle

    # Check the GRIB header
    data_handle.open()
    assert data_handle.read(4) == b"GRIB"
    assert data_handle.read(4) != b"GRIB"


def test_datahandle_readinto(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        assert data_handle

        chunk_buf = bytearray(4)
        mv = memoryview(chunk_buf)
        data_handle.readinto(mv)

        assert chunk_buf == b"GRIB"
        assert data_handle.read(4) != b"GRIB"


def test_datahandle_readinto_non_opened(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)
    assert data_handle

    chunk_buf = bytearray(4)
    mv = memoryview(chunk_buf)

    with pytest.raises(RuntimeError):
        data_handle.readinto(mv)


def test_datahandle_read_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)

    # Check the GRIB header
    assert data_handle
    assert data_handle.readall()[0:4] == b"GRIB"
    data_handle.close()


def test_datahandle_cmp_read_read_all(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)
    data_handle_2 = fdb.retrieve(selection)

    assert data_handle
    assert data_handle_2

    data_handle_2.open()

    # Read all the data from data_handle
    data = data_handle.readall()
    data_2 = data_handle_2.read(len(data))

    assert data == data_2

    data_handle.open()
    print(data_handle.read(4))
    print(data_handle_2.read(4))


def test_datahandle_not_opened_before_read_context_manager(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        # Check the GRIB header
        assert data_handle
        assert data_handle.read(4) == b"GRIB"


def test_datahandle_not_opened_before_read_all_context_manager(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        assert data_handle
        assert data_handle.readall()


def test_datahandle_size(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        assert data_handle.size() == 10732


def test_datahandle_size_without_open(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)
    assert data_handle.size() == 10732


def test_datahandle_readinto_shutil_copyfileobj_cmp(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    dst_read_into = io.BytesIO(b"")

    with fdb.retrieve(selection) as data_handle:
        assert data_handle
        print("Reading with shutil")
        shutil.copyfileobj(data_handle, dst_read_into)

        # reset position in file
        dst_read_into.seek(0)

    content_read_into = dst_read_into.read()

    with fdb.retrieve(selection) as data_handle:
        assert data_handle

        content_read_all = data_handle.readall()

        assert len(content_read_all) == len(content_read_into)
        assert content_read_all == content_read_into


def test_datahandle_readinto_shutil_copyfileobj_tmp_file(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with tempfile.TemporaryFile() as out:
        with fdb.retrieve(selection) as data_handle:
            assert data_handle
            shutil.copyfileobj(data_handle, out)
            out.flush()

        out.seek(0)
        content_read_into = out.read(-1)

        with fdb.retrieve(selection) as data_handle:
            assert data_handle

            content_read_all = data_handle.readall()

            assert len(content_read_all) == len(content_read_into)
            assert content_read_all == content_read_into


def test_data_handle_readinto_attr(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        assert hasattr(data_handle, "readinto")


def test_data_handle_read_minus_one(read_only_fdb_setup):
    fdb_config_path = read_only_fdb_setup

    assert fdb_config_path

    fdb = FDB(fdb_config_path)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    content_read_minus_one = None
    content_read_all = None

    with fdb.retrieve(selection) as data_handle:
        content_read_minus_one = data_handle.read(-1)

    content_read_all = data_handle.readall()

    assert content_read_all
    assert content_read_minus_one
    assert len(content_read_minus_one) > 0
    assert len(content_read_all) > 0
    assert content_read_all == content_read_minus_one



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from zarr.abc.store import Store
from pathlib import Path

from z3fdb._internal.zarr import FdbZarrStore, FdbZarrArray, FdbSource
from pychunked_data_view import (
    ChunkedDataViewBuilder,
    AxisDefinition,
    ExtractorType,
)


class SimpleStoreBuilder:
    """Builder to create a Zarr store with FDB backing.

    This builder will create a Zarr store with a Zarr Array at its root ("/")
    containing the data from your MARS request(s).

    Args:
        fdb_config_file: Optional path to FDB config file. If not set normal
            FDB config file resolution is applied.
    """

    def __init__(self, fdb_config_file: Path | None = None):
        self._builder = ChunkedDataViewBuilder(fdb_config_file)

    def add_part(
        self,
        mars_request_key_values: str,
        axes: list[AxisDefinition],
        extractor_type: ExtractorType,
    ) -> None:
        """Add a MARS request to the view.

        Args:
            mars_request_key_values(str):
                A string with fully spefified MARS key:values pairs.

                For example:
                    type=an,
                    class=ea,
                    domain=g,
                    expver=0001,
                    stream=oper,
                    date=2020-01-01/to/2020-01-04,
                    levtype=sfc,
                    step=0,
                    param=167/131/132,
                    time=0/to/21/by/3

            axes(:obj:`list` of :obj:`AxisDefinition`):  List of
                AxisDefinitions that describe how axis in the MARS request are
                mapped to axis in the Zarr array.
            extractor_type: Defines how to extract data from FDB. Currently
                only ExtractorType.GRIB is supported.
        """
        self._builder.add_part(mars_request_key_values, axes, extractor_type)

    def extend_on_axis(self, axis: int) -> None:
        """Defines the extension axis when multiple parts are added.

        Args:
            axis(int): Index of the axis that is extendet when multiple parts
                have been added.

        """
        self._builder.extend_on_axis(axis)

    def build(self) -> Store:
        """Build the store from the inputs.

        Raises:
            Z3fdbError if store cannot be created.

        """
        return FdbZarrStore(
            FdbZarrArray(
                datasource=FdbSource(self._builder.build()),
            )
        )



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

# libfdb5.so and ependencies have to be loaded prior to importing
# pychunked_data_view
import findlibs

findlibs.load("fdb5")

from pychunked_data_view.chunked_data_view import (  # noqa: E402
    AxisDefinition,
    ChunkedDataView,
    ChunkedDataViewBuilder,
    Chunking,
    ExtractorType,
)

__all__ = [
    "AxisDefinition",
    "ChunkedDataView",
    "ChunkedDataViewBuilder",
    "Chunking",
    "ExtractorType",
]



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from pyfdb.pyfdb_iterator import (
    ControlElement,
    IndexAxis,
    ListElement,
    PurgeElement,
    StatsElement,
    StatusElement,
    WipeElement,
)
from pyfdb.pyfdb_type import DataHandle

classes = [
    IndexAxis,
    ListElement,
    StatusElement,
    PurgeElement,
    StatsElement,
    ControlElement,
    WipeElement,
    DataHandle,
]


@pytest.mark.parametrize("cls", classes)
def test_initialization_raises_error(cls):
    """Test for creation of internal data types with certain elements."""
    with pytest.raises(TypeError):
        cls(None)
